"""Mixins for LocalSite-related views and forms."""

import logging

from django.contrib.auth.models import User
from django.core.exceptions import FieldDoesNotExist
from django.utils.decorators import method_decorator
from django.utils.translation import gettext as _
from djblets.db.query import get_object_or_none
from djblets.forms.fields import ConditionsField

from reviewboard.admin.form_widgets import RelatedObjectWidget
from reviewboard.site.decorators import check_local_site_access
from reviewboard.site.models import LocalSite


logger = logging.getLogger(__name__)


class CheckLocalSiteAccessViewMixin(object):
    """Generic view mixin to check if a user has access to the Local Site.

    It's important to note that this does not check for login access.
    This is just a convenience around using the
    :py:func:`@check_local_site_access
    <reviewboard.site.decorators.check_local_site_access>` decorator for
    generic views.

    The :py:attr:`local_site` attribute will be set on the class for use in
    the view.

    Attributes:
        local_site (reviewboard.site.models.LocalSite):
            The Local Site being accessed, or ``None``.
    """

    @method_decorator(check_local_site_access)
    def dispatch(self, request, local_site=None, *args, **kwargs):
        """Dispatch a HTTP request to the right handler.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            local_site (reviewboard.site.models.LocalSite, optional):
                The Local Site being accessed, if any.

            *args (tuple):
                Positional arguments to pass to the handler.

            **kwargs (tuple):
                Keyword arguments to pass to the handler.

                These will be arguments provided by the URL pattern.

        Returns:
            django.http.HttpResponse:
            The resulting HTTP response to send to the client.
        """
        self.local_site = local_site

        return super(CheckLocalSiteAccessViewMixin, self).dispatch(
            request, *args, **kwargs)


class LocalSiteAwareModelFormMixin(object):
    """Mixin for model forms that associate with Local Sites.

    This mixin allows model forms to be bound to a Local Site, which will
    then limit all relation fields to objects bound to the same site. This
    allows the construction of forms that could be modified by a Local Site
    administrator without risk of data outside the Local Site being used. The
    bound Local Site will always be forced, and the field will not appear on
    the form.

    When not bound to a Local Site, relations across all Local Sites (and
    those not associated with a Local Site) are allowed, but if a Local Site
    is being assigned when posting to the form, all related objects will be
    validated against that Local Site.

    Attributes:
        cur_local_site (reviewboard.site.models.LocalSite):
            The Local Site that's either been bound to the form or provided
            in posted data. This will be ``None`` if no Local Site is being
            used.

        limited_to_local_site (reviewboard.site.models.LocalSite):
            The Local Site bound to the form. This will be ``None`` if no
            Local Site is bound.

        request (django.http.HttpRequest):
            The HTTP request provided to the form. This may be ``None`` if no
            request was provided.
    """

    #: The name of the Local Site field on the form.
    local_site_field_name = 'local_site'

    #: Whether the form needs the 'request' argument.
    form_needs_request = False

    def __init__(self, data=None, initial={}, request=None, *args, **kwargs):
        """Initialize the form.

        Args:
            data (dict, optional):
                Posted data for the form.

            initial (dict, optional):
                Initial data for the form.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client. This is used for logging
                of access errors, and will be passed to the underlying form
                if :py:attr:`form_needs_request` is ``True``.

            *args (tuple):
                Positional arguments to pass to the parent form.

            **kwargs (dict):
                Keyword arguments to pass to the parent form.

        Keyword Args:
            limit_to_local_site (reviewboard.site.models.LocalSite, optional):
                A specific Local Site to bind the form to.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

        Raises:
            ValueError:
                An object instance was provided to the form that isn't
                compatible with the
                :py:class:`~reviewboard.site.models.LocalSite` provided in
                ``limit_to_local_site``.
        """
        local_site = kwargs.pop('limit_to_local_site', None)

        if self.form_needs_request:
            kwargs['request'] = request

        self.limited_to_local_site = local_site

        local_site_field_name = self.local_site_field_name

        if local_site is None:
            local_site_id = None

            if data and data.get(local_site_field_name):
                try:
                    local_site_id = int(data[local_site_field_name])
                except ValueError:
                    local_site_id = None

            if local_site_id is not None:
                local_site = get_object_or_none(LocalSite, pk=local_site_id)
            elif initial and initial.get(local_site_field_name):
                local_site = initial[local_site_field_name]

        if data is not None and local_site:
            # Always use the assigned Local Site in the model data. We
            # don't want to allow this to ever be overridden. We'll delete
            # here and then assign after the constructor does its thing.
            data = data.copy()
            data[local_site_field_name] = local_site.pk

        self.cur_local_site = local_site

        super(LocalSiteAwareModelFormMixin, self).__init__(
            data=data,
            initial=initial,
            *args,
            **kwargs)

        # Prepare to patch up some fields. We have a few special types we'll
        # be dealing with.
        self._conditions_fields = []
        self._related_obj_fields = []
        self._queryset_fields = []
        self._patched_local_sites = False

        for field_name, field in self.fields.items():
            if isinstance(field.widget, RelatedObjectWidget):
                self._related_obj_fields.append(field)
            elif isinstance(field, ConditionsField):
                self._conditions_fields.append(field)
                field.choice_kwargs['request'] = request

            if getattr(field, 'queryset', None) is not None:
                self._queryset_fields.append(field)

        if self.limited_to_local_site is not None:
            if self.instance is not None:
                if self.instance.pk is None:
                    # This is a new instance, so force its Local Site now.
                    self.instance.local_site = local_site
                elif self.instance.local_site != local_site:
                    # Something went very wrong, and an instance is now in our
                    # form that isn't part of this Local Site. Log this and
                    # bail out now.
                    logger.error('Attempted to pass instance %r with '
                                 'LocalSite "%s" to form %r. Only LocalSite '
                                 '"%s" is permitted.',
                                 self.instance,
                                 self.instance.local_site,
                                 self.__class__,
                                 local_site,
                                 extra={'request': request})

                    raise ValueError(
                        _('The provided instance is not associated with a '
                          'LocalSite compatible with this form. Please '
                          'contact support.'))

            # We never want to show a "Local Site" field, so let's get rid of
            # it.
            del self.fields[local_site_field_name]

            # Go through the fields and widgets and start limiting querysets
            # and other choices.
            local_site_name = local_site.name

            for field in self._related_obj_fields:
                field.widget.local_site_name = local_site_name

            for field in self._conditions_fields:
                field.choice_kwargs['local_site'] = local_site

            for field in self._queryset_fields:
                self._patch_field_local_site_queryset(field, local_site)

            self._patched_local_sites = True

    def full_clean(self):
        """Perform a full clean the form.

        This wraps the typical form cleaning process by first ensuring that
        all relation fields limit their choices to objects on the bound
        Local Site (if one is set), allowing validation to work naturally on
        each field. It then invokes the standard form cleaning logic, and
        then restores the choices.

        Returns:
            dict:
            The cleaned data from the form.

        Raises:
            django.core.exceptions.ValidationError:
                The form failed to validate.
        """
        local_site = self.cur_local_site

        assert (self.limited_to_local_site is None or
                self.limited_to_local_site == local_site)

        if local_site is None:
            local_site_name = None
        else:
            local_site_name = local_site.name

        # Go through the fields and widgets and start limiting querysets
        # and other choices.
        old_values = {}

        # Only perform this work if we didn't do it during construction.
        if not self._patched_local_sites:
            for field in self._related_obj_fields:
                old_values[field.widget] = ('local_site_name',
                                            field.widget.local_site_name)
                field.widget.local_site_name = local_site_name

            for field in self._conditions_fields:
                old_values[field] = ('choice_kwargs',
                                     field.choice_kwargs.copy())
                field.choice_kwargs['local_site'] = local_site

            for field in self._queryset_fields:
                old_queryset = self._patch_field_local_site_queryset(
                    field, local_site)

                if old_queryset is not None:
                    old_values[field] = ('queryset', old_queryset)

        try:
            return super(LocalSiteAwareModelFormMixin, self).full_clean()
        finally:
            # Restore the values so that the original options are available
            # if the form is shown again (due to errors).
            for obj, (attr, value) in old_values.items():
                setattr(obj, attr, value)

    def _clean_fields(self):
        """Clean the fields on the form.

        This overrides the standard form field cleaning logic to ensure that
        a Local Site is available in the cleaned data if binding to a Local
        Site. This is required because in this situation, the field will not
        be present on the form, even though other clean methods may require
        it.
        """
        if self.limited_to_local_site is not None:
            self.cleaned_data[self.local_site_field_name] = \
                self.limited_to_local_site

        super(LocalSiteAwareModelFormMixin, self)._clean_fields()

    def _patch_field_local_site_queryset(self, field, local_site):
        """Patch the queryset on a field to be bound to a Local Site.

        The field will only be patched if it contains a ``queryset`` attribute
        and its related model type contains a known attribute for associating
        with a :py:class:`~reviewboard.site.models.LocalSite`.

        Note that if the queryset matches users, then we only perform the
        patching if the model instance will have an associated
        :py:class:`~reviewboard.site.models.LocalSite`. This allows any user
        (whether or not they're on a Local Site) to be associated with
        instances not on a Local Site, preserving the behavior we've had for
        many years.

        If the queryset matches any other type of model, then we're more
        strict, requiring the object's Local Site value to be the same as
        what's being passed in to this method.

        Args:
            field (django.forms.Field):
                The field to patch.

            local_site (reviewboard.site.models.LocalSite):
                The Local Site to bind the queryset to.

        Returns:
            django.db.models.query.QuerySet:
            The old queryset for the field. This will be ``None`` if the
            queryset was not patched.
        """
        queryset = getattr(field, 'queryset', None)
        old_queryset = None

        if (queryset is not None and
            (local_site is not None or
             not issubclass(queryset.model, User))):
            local_site_field_name = \
                self._get_rel_local_site_field_name(queryset.model)

            if local_site_field_name:
                old_queryset = queryset
                field.queryset = queryset.filter(**{
                    local_site_field_name: local_site,
                })

        return old_queryset

    def _get_rel_local_site_field_name(self, model):
        """Return the Local Site field name on a model.

        This checks for the presence of a ``local_site`` or ``local_site_set``
        field.

        Args:
            model (django.db.models.Model):
                The model containing the field.

        Returns:
            unicode:
            The name of the Local Site field, if found. If a field with a
            supported name is not present, ``None`` will be returned.
        """
        meta = model._meta

        for local_site_field_name in ('local_site', 'local_site_set'):
            try:
                meta.get_field(local_site_field_name)

                return local_site_field_name
            except FieldDoesNotExist:
                continue

        return None



"""Unit tests for reviewboard.admin.forms.change_form.ChangeFormRow."""

from django.contrib.auth.models import User

from reviewboard.admin import admin_site
from reviewboard.admin.forms.change_form import (ChangeFormField,
                                                 ChangeFormRow)
from reviewboard.testing.testcase import TestCase


class ChangeFormRowTests(TestCase):
    """Unit tests for ChangeFormRow."""

    def setUp(self):
        super(ChangeFormRowTests, self).setUp()

        request = self.create_http_request()
        model_admin = admin_site.get_model_admin(User)

        self.form = model_admin.get_form(request)()

    def test_classes(self):
        """Testing ChangeFormRow.classes"""
        row = ChangeFormRow(self.form,
                            field='username')

        self.assertEqual(row.classes, 'rb-c-form-row field-username')

    def test_iter(self):
        """Testing ChangeFormRow.__iter__"""
        row = ChangeFormRow(self.form,
                            field='username')

        fields = list(row)
        self.assertEqual(len(fields), 1)

        field = fields[0]
        self.assertIsInstance(field, ChangeFormField)
        self.assertEqual(field.field.name, 'username')



"""Models for reviews."""

from __future__ import annotations

import logging
from typing import ClassVar, Optional, TYPE_CHECKING

from django.contrib.auth.models import User
from django.db import models
from django.db.models import Q
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import gettext, gettext_lazy as _
from djblets.db.fields import CounterField, JSONField
from djblets.siteconfig.models import SiteConfiguration

from reviewboard.diffviewer.models import DiffSet
from reviewboard.reviews.errors import PublishError, RevokeShipItError
from reviewboard.reviews.managers import ReviewManager
from reviewboard.reviews.models.base_comment import BaseComment
from reviewboard.reviews.models.diff_comment import Comment
from reviewboard.reviews.models.file_attachment_comment import \
    FileAttachmentComment
from reviewboard.reviews.models.general_comment import GeneralComment
from reviewboard.reviews.models.review_request import (ReviewRequest,
                                                       fetch_issue_counts)
from reviewboard.reviews.models.screenshot_comment import ScreenshotComment
from reviewboard.reviews.signals import (reply_publishing, reply_published,
                                         review_publishing, review_published,
                                         review_ship_it_revoking,
                                         review_ship_it_revoked)

if TYPE_CHECKING:
    from django.http import HttpRequest


logger = logging.getLogger(__name__)


class Review(models.Model):
    """A review of a review request."""

    # Constants used in e-mails when a review contains a Ship It designation.
    # These are explicitly not marked for localization to prevent taking the
    # submitting user's local into account when generating the e-mail.
    SHIP_IT_TEXT = 'Ship It!'
    REVOKED_SHIP_IT_TEXT = '~~Ship It!~~'
    FIX_IT_THEN_SHIP_IT_TEXT = 'Fix it, then Ship it!'

    review_request = models.ForeignKey(ReviewRequest,
                                       on_delete=models.CASCADE,
                                       related_name='reviews',
                                       verbose_name=_('review request'))
    user = models.ForeignKey(User,
                             on_delete=models.CASCADE,
                             verbose_name=_('user'),
                             related_name='reviews')
    timestamp = models.DateTimeField(_('timestamp'), default=timezone.now)
    public = models.BooleanField(_('public'), default=False)
    ship_it = models.BooleanField(
        _('ship it'),
        default=False,
        help_text=_('Indicates whether the reviewer thinks this code is '
                    'ready to ship.'))
    base_reply_to = models.ForeignKey(
        'self',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        related_name='replies',
        verbose_name=_('Base reply to'),
        help_text=_('The top-most review in the discussion thread for '
                    'this review reply.'))
    email_message_id = models.CharField(_('e-mail message ID'), max_length=255,
                                        blank=True, null=True)
    time_emailed = models.DateTimeField(_('time e-mailed'), null=True,
                                        default=None, blank=True)

    body_top = models.TextField(
        _('body (top)'),
        blank=True,
        help_text=_('The review text shown above the diff and screenshot '
                    'comments.'))
    body_top_rich_text = models.BooleanField(
        _('body (top) in rich text'),
        default=False)

    body_bottom = models.TextField(
        _('body (bottom)'),
        blank=True,
        help_text=_('The review text shown below the diff and screenshot '
                    'comments.'))
    body_bottom_rich_text = models.BooleanField(
        _('body (bottom) in rich text'),
        default=False)

    body_top_reply_to = models.ForeignKey(
        'self',
        on_delete=models.CASCADE,
        blank=True, null=True,
        related_name='body_top_replies',
        verbose_name=_('body (top) reply to'),
        help_text=_('The review that the body (top) field is in reply to.'))
    body_bottom_reply_to = models.ForeignKey(
        'self',
        on_delete=models.CASCADE,
        blank=True, null=True,
        related_name='body_bottom_replies',
        verbose_name=_('body (bottom) reply to'),
        help_text=_('The review that the body (bottom) field is in reply to.'))

    comments = models.ManyToManyField(Comment, verbose_name=_('comments'),
                                      related_name='review', blank=True)
    screenshot_comments = models.ManyToManyField(
        ScreenshotComment,
        verbose_name=_('screenshot comments'),
        related_name='review',
        blank=True)
    file_attachment_comments = models.ManyToManyField(
        FileAttachmentComment,
        verbose_name=_('file attachment comments'),
        related_name='review',
        blank=True)
    general_comments = models.ManyToManyField(
        GeneralComment,
        verbose_name=_('general comments'),
        related_name='review',
        blank=True)

    extra_data = JSONField(null=True)

    # Deprecated and no longer used for new reviews as of 2.0.9.
    rich_text = models.BooleanField(_('rich text'), default=False)

    # XXX Deprecated. This will be removed in a future release.
    reviewed_diffset = models.ForeignKey(
        DiffSet,
        on_delete=models.CASCADE,
        verbose_name='Reviewed Diff',
        blank=True, null=True,
        help_text=_('This field is unused and will be removed in a future '
                    'version.'))

    # Set this up with a ReviewManager to help prevent race conditions and
    # to fix duplicate reviews.
    objects: ClassVar[ReviewManager] = ReviewManager()

    @cached_property
    def ship_it_only(self):
        """Return if the review only contains a "Ship It!".

        Returns:
            bool: ``True`` if the review is only a "Ship It!" and ``False``
            otherwise.
        """
        return (self.ship_it and
                (not self.body_top or
                 self.body_top == Review.SHIP_IT_TEXT) and
                not (self.body_bottom or
                     self.has_comments(only_issues=False)))

    def can_user_revoke_ship_it(self, user):
        """Return whether a given user can revoke a Ship It.

        Args:
            user (django.contrib.auth.models.User):
                The user to check permissions for.

        Returns:
            bool:
            ``True`` if the user has permissions to revoke a Ship It.
            ``False`` if they don't.
        """
        return (user.is_authenticated and
                self.public and
                (user.pk == self.user_id or
                 user.is_superuser or
                 (self.review_request.local_site and
                  self.review_request.local_site.admins.filter(
                      pk=user.pk).exists())) and
                self.review_request.is_accessible_by(user))

    def revoke_ship_it(self, user):
        """Revoke the Ship It status on this review.

        The Ship It status will be removed, and the
        :py:data:`ReviewRequest.shipit_count
        <reviewboard.reviews.models.review_request.ReviewRequest.shipit_count>`
        counter will be decremented.

        If the :py:attr:`body_top` text is equal to :py:attr:`SHIP_IT_TEXT`,
        then it will replaced with :py:attr:`REVOKED_SHIP_IT_TEXT`.

        Callers are responsible for checking whether the user has permission
        to revoke Ship Its by using :py:meth:`can_user_revoke_ship_it`.

        Raises:
            reviewboard.reviews.errors.RevokeShipItError:
                The Ship It could not be revoked. Details will be in the
                error message.
        """
        if not self.ship_it:
            raise RevokeShipItError('This review is not marked Ship It!')

        # This may raise a RevokeShipItError.
        try:
            review_ship_it_revoking.send(sender=self.__class__,
                                         user=user,
                                         review=self)
        except RevokeShipItError:
            raise
        except Exception as e:
            logger.exception('Unexpected error notifying listeners before '
                             'revoking a Ship It for review ID=%d: %s',
                             self.pk, e)
            raise RevokeShipItError(e)

        if self.extra_data is None:
            self.extra_data = {}

        self.extra_data['revoked_ship_it'] = True
        self.ship_it = False

        update_fields = ['extra_data', 'ship_it']

        if self.body_top == self.SHIP_IT_TEXT:
            self.body_top = self.REVOKED_SHIP_IT_TEXT
            self.body_top_rich_text = True
            update_fields += ['body_top', 'body_top_rich_text']

        self.save(update_fields=update_fields)

        self.review_request.decrement_shipit_count()
        self.review_request.last_review_activity_timestamp = timezone.now()
        self.review_request.save(
            update_fields=['last_review_activity_timestamp'])

        try:
            review_ship_it_revoked.send(sender=self.__class__,
                                        user=user,
                                        review=self)
        except Exception as e:
            logger.exception('Unexpected error notifying listeners after '
                             'revoking a Ship It for review ID=%d: %s',
                             self.pk, e)

    @cached_property
    def all_participants(self):
        """Return all participants in the review's discussion.

        This will always contain the user who filed the review, plus every user
        who has published a reply to the review.

        The result is cached. Repeated calls will return the same result.

        Returns:
            set of django.contrib.auth.models.User:
            The users who participated in the discussion.
        """
        user_ids = (
            self.replies
            .filter(public=True)
            .values_list('user_id', flat=True)
        )
        user_id_lookup = set(user_ids) - {self.user.pk}
        users = {self.user}

        if user_id_lookup:
            users.update(User.objects.filter(pk__in=user_id_lookup))

        return users

    def is_accessible_by(self, user):
        """Returns whether the user can access this review."""
        return ((self.public or
                 user.is_superuser or
                 self.user_id == user.pk) and
                self.review_request.is_accessible_by(user))

    def is_mutable_by(self, user):
        """Returns whether the user can modify this review."""
        return ((not self.public and
                 (user.is_superuser or
                  self.user_id == user.pk)) and
                self.review_request.is_accessible_by(user))

    def __str__(self):
        return "Review of '%s'" % self.review_request

    def is_reply(self):
        """Returns whether or not this review is a reply to another review."""
        return self.base_reply_to_id is not None
    is_reply.boolean = True

    def is_new_for_user(self, user, last_visited):
        """Return whether this review is new for a user.

        The review is considered new if their last visited time is older than
        the review's published timestamp and the user is not the one who
        created the review.

        Args:
            user (django.contrib.auth.models.User):
                The user accessing the review.

            last_visited (datetime.datetime):
                The last time the user accessed a page where the review would
                be shown.

        Returns:
            bool:
            ``True`` if the review is new to this user. ``False`` if it's older
            than the last visited time or the user created it.
        """
        return user.pk != self.user_id and last_visited < self.timestamp

    def public_replies(self):
        """Returns a list of public replies to this review."""
        return self.replies.filter(public=True)

    def public_body_top_replies(self, user=None):
        """Returns a list of public replies to this review's body top."""
        if hasattr(self, '_body_top_replies'):
            return self._body_top_replies
        else:
            q = Q(public=True)

            if user and user.is_authenticated:
                q = q | Q(user=user)

            return self.body_top_replies.filter(q).order_by('timestamp')

    def public_body_bottom_replies(self, user=None):
        """Returns a list of public replies to this review's body bottom."""
        if hasattr(self, '_body_bottom_replies'):
            return self._body_bottom_replies
        else:
            q = Q(public=True)

            if user and user.is_authenticated:
                q = q | Q(user=user)

            return self.body_bottom_replies.filter(q).order_by('timestamp')

    def get_pending_reply(self, user):
        """Return the pending reply owned by the specified user.

        Args:
            user (django.contrib.auth.models.User):
                The user to find the reply for.

        Returns:
            reviewboard.reviews.models.review.Review:
            The pending reply object, if present. ``None`` if there is no
            pending reply.
        """
        return Review.objects.get_pending_reply(self, user)

    def save(self, **kwargs):
        if ('update_fields' not in kwargs or
            'timestamp' in kwargs['update_fields']):
            self.timestamp = timezone.now()

        super(Review, self).save(**kwargs)

    def can_publish(
        self,
        review_request_will_publish: bool = False,
    ) -> tuple[bool, Optional[str]]:
        """Check if this review can be published.

        Args:
            review_request_will_publish (bool, optional):
                Whether the review request is also being published.

        Returns:
            tuple:
            A two-tuple describing whether the review can be published, and if
            not, why not.

            Tuple:
                1 (bool):
                    Whether the review can be published.

                2 (str):
                    An error message describing why the review cannot be
                    published.
        """
        if not review_request_will_publish:
            if not self.review_request.public:
                return False, gettext(
                    'This review cannot be published until the review request '
                    'is published.')

            diff_comments = list(self.comments.all())
            file_attachment_comments = list(
                self.file_attachment_comments.all())

            if (not self.body_bottom and
                not self.body_top and
                not self.ship_it and
                not diff_comments and
                not file_attachment_comments and
                not self.screenshot_comments.exists() and
                not self.general_comments.exists()):
                return False, gettext(
                    'This review cannot be published, because it is empty.')

            for comment in diff_comments:
                if not comment.diff_is_public():
                    return False, gettext(
                        'This review cannot be published, because it includes '
                        'a comment on a diff which has not yet been '
                        'published.')

            for comment in file_attachment_comments:
                if not comment.attachment_is_public():
                    return False, gettext(
                        'This review cannot be published, because it includes '
                        'a comment on a file attachment which has not yet '
                        'been published.')

        siteconfig = SiteConfiguration.objects.get_current()

        if (self.ship_it and
            self.user_id == self.review_request.submitter_id and
            not siteconfig.get('reviews_allow_self_shipit')):
            return False, gettext(
                'You cannot mark "Ship It!" on your own review request.')

        return True, None

    def publish(
        self,
        user: Optional[User] = None,
        trivial: bool = False,
        to_owner_only: bool = False,
        request: Optional[HttpRequest] = None,
        review_request_will_publish: bool = False,
        *args,
        **kwargs,
    ) -> None:
        """Publish this review.

        This will make the review public and update the timestamps of all
        contained comments.

        Args:
            user (django.contrib.auth.models.User, optional):
                The user publishing the review.

            trivial (bool, optional):
                Whether to skip any e-mail notifications.

            to_owner_only (bool, optional):
                Whether to address notifications only to the owner of the
                review request.

            request (djang.http.HttpRequest, optional):
                The HTTP request.

            review_request_will_publish (bool, optional):
                Whether the review request is also being published.

                Version Added:
                    6.0

            *args (tuple):
                Positional arguments for future expansion.

            **kwargs (dict):
                Keyword arguments for future expansion.
        """
        can_publish, err = self.can_publish(
            review_request_will_publish=review_request_will_publish)

        if not can_publish:
            raise PublishError(err)

        if not user:
            user = self.user

        self.public = True

        if self.is_reply():
            reply_publishing.send(sender=self.__class__, user=user, reply=self)
        else:
            review_publishing.send(sender=self.__class__, user=user,
                                   review=self)

        self.save()

        self.comments.update(timestamp=self.timestamp)
        self.screenshot_comments.update(timestamp=self.timestamp)
        self.file_attachment_comments.update(timestamp=self.timestamp)
        self.general_comments.update(timestamp=self.timestamp)

        # Update the last_updated timestamp and the last review activity
        # timestamp on the review request.
        self.review_request.last_review_activity_timestamp = self.timestamp
        self.review_request.last_updated = self.timestamp
        self.review_request.save(update_fields=(
            'last_review_activity_timestamp', 'last_updated'))

        if self.is_reply():
            reply_published.send(sender=self.__class__,
                                 user=user, reply=self, trivial=trivial)
        else:
            issue_counts = fetch_issue_counts(self.review_request,
                                              Q(pk=self.pk))

            # Since we're publishing the review, all filed issues should be
            # open.
            assert issue_counts[BaseComment.RESOLVED] == 0
            assert issue_counts[BaseComment.DROPPED] == 0
            assert issue_counts[BaseComment.VERIFYING_RESOLVED] == 0
            assert issue_counts[BaseComment.VERIFYING_DROPPED] == 0

            if self.ship_it:
                ship_it_value = 1
            else:
                ship_it_value = 0

            # Atomically update the issue count and Ship It count.
            CounterField.increment_many(
                self.review_request,
                {
                    'issue_open_count': issue_counts[BaseComment.OPEN],
                    'issue_dropped_count': 0,
                    'issue_resolved_count': 0,
                    'issue_verifying_count': 0,
                    'shipit_count': ship_it_value,
                })

            review_published.send(sender=self.__class__,
                                  user=user, review=self,
                                  to_owner_only=to_owner_only,
                                  trivial=trivial,
                                  request=request)

    def delete(self):
        """Deletes this review.

        This will enforce that all contained comments are also deleted.
        """
        self.comments.all().delete()
        self.screenshot_comments.all().delete()
        self.file_attachment_comments.all().delete()
        self.general_comments.all().delete()

        super(Review, self).delete()

    def get_absolute_url(self):
        return "%s#review%s" % (self.review_request.get_absolute_url(),
                                self.pk)

    def get_all_comments(self, **kwargs):
        """Return a list of all contained comments of all types."""
        return (list(self.comments.filter(**kwargs)) +
                list(self.screenshot_comments.filter(**kwargs)) +
                list(self.file_attachment_comments.filter(**kwargs)) +
                list(self.general_comments.filter(**kwargs)))

    def has_comments(self, only_issues=False):
        """Return whether the review contains any comments/issues.

        Args:
            only_issues (bool, optional):
                Whether or not to check for comments where ``issue_opened`` is
                ``True``. ``True`` to check for issues, or ``False`` to check
                for comments only. Defaults to ``False``.

        Returns:
            bool:
            ``True`` if the review contains any comments/issues and
            ``False`` otherwise.
        """
        qs = [
            self.comments,
            self.file_attachment_comments,
            self.screenshot_comments,
            self.general_comments,
        ]

        if only_issues:
            qs = [
                q.filter(issue_opened=True)
                for q in qs
            ]

        return any(q.exists() for q in qs)

    class Meta:
        app_label = 'reviews'
        db_table = 'reviews_review'
        ordering = ['timestamp']
        get_latest_by = 'timestamp'
        verbose_name = _('Review')
        verbose_name_plural = _('Reviews')



"""Unit tests for reviewboard.datagrids.columns.ReviewRequestIDColumn.

Version Added:
    8.0
"""

from __future__ import annotations

from django.utils.safestring import SafeString
from djblets.testing.decorators import add_fixtures

from reviewboard.datagrids.columns import ReviewRequestIDColumn
from reviewboard.datagrids.tests.base import BaseColumnTestCase


class ReviewRequestIDColumnTests(BaseColumnTestCase):
    """Unit tests for reviewboard.datagrids.columns.ReviewRequestIDColumn.

    Version Added:
        8.0
    """

    column = ReviewRequestIDColumn()

    def test_render_data(self) -> None:
        """Testing ReviewRequestIDColumn.render_data"""
        review_request = self.create_review_request(publish=True)

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value, '1')

    @add_fixtures(['test_site'])
    def test_render_data_with_local_site(self) -> None:
        """Testing ReviewRequestIDColumn.render_data with LocalSite"""
        review_request = self.create_review_request(
            publish=True,
            with_local_site=True)

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value, '1001')

    def test_to_json(self) -> None:
        """Testing ReviewRequestIDColumn.to_json"""
        review_request = self.create_review_request(publish=True)

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            1)

    @add_fixtures(['test_site'])
    def test_to_json_with_local_site(self) -> None:
        """Testing ReviewRequestIDColumn.to_json with LocalSite"""
        review_request = self.create_review_request(
            publish=True,
            with_local_site=True)

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            1001)



"""Administration form for privacy settings."""

from django import forms
from django.utils.translation import gettext_lazy as _
from djblets.avatars.registry import AvatarServiceRegistry
from djblets.siteconfig.forms import SiteSettingsForm


class PrivacySettingsForm(SiteSettingsForm):
    """Site-wide user privacy settings for Review Board."""

    terms_of_service_url = forms.URLField(
        label=_('Terms of service URL'),
        required=False,
        help_text=_('URL to your terms of service. This will be displayed on '
                    'the My Account page and during login and registration.'),
        widget=forms.widgets.URLInput(attrs={
            'size': 80,
        }))

    privacy_policy_url = forms.URLField(
        label=_('Privacy policy URL'),
        required=False,
        help_text=_('URL to your privacy policy. This will be displayed on '
                    'the My Account page and during login and registration.'),
        widget=forms.widgets.URLInput(attrs={
            'size': 80,
        }))

    privacy_info_html = forms.CharField(
        label=_('Privacy information'),
        required=False,
        help_text=_('A description of the privacy guarantees for users on '
                    'this server. This will be displayed on the My Account '
                    '-> Your Privacy page. HTML is allowed.'),
        widget=forms.widgets.Textarea(attrs={
            'cols': 60,
        }))

    privacy_enable_user_consent = forms.BooleanField(
        label=_('Require consent for usage of personal information'),
        required=False,
        help_text=_('Require consent from users when using their personally '
                    'identifiable information (usernames, e-mail addresses, '
                    'etc.) for when talking to third-party services, like '
                    'Gravatar. This is required for EU GDPR compliance.'))

    def save(self):
        """Save the privacy settings form.

        This will write the new configuration to the database.
        """
        self.siteconfig.set(AvatarServiceRegistry.ENABLE_CONSENT_CHECKS,
                            self.cleaned_data['privacy_enable_user_consent'])

        super(PrivacySettingsForm, self).save()

    class Meta:
        title = _('User Privacy Settings')
        fieldsets = (
            {
                'classes': ('wide',),
                'fields': (
                    'terms_of_service_url',
                    'privacy_policy_url',
                    'privacy_info_html',
                    'privacy_enable_user_consent',
                ),
            },
        )



"""Hooks for adding to the "My Account" page."""

from __future__ import annotations

from djblets.extensions.hooks import (BaseRegistryMultiItemHook,
                                      ExtensionHook,
                                      ExtensionHookPoint)

from reviewboard.accounts.pages import AccountPage


class AccountPagesHook(BaseRegistryMultiItemHook,
                       metaclass=ExtensionHookPoint):
    """A hook for adding new pages to the My Account page.

    A page can contain one or more forms or even a custom template allowing
    for configuration of an extension.

    This takes a list of AccountPage classes as parameters, which it will
    later instantiate as necessary. Each page can be pre-populated with
    one or more custom AccountPageForm classes.
    """

    registry = AccountPage.registry

    def initialize(self, page_classes):
        """Initialize the hook.

        This will register each of the provided account page classes.

        Args:
            page_classes (list of type):
                The list of page classes to register. Each must be a subclass
                of :py:class:`~reviewboard.accounts.pages.AccountPage`.
        """
        super(AccountPagesHook, self).initialize(page_classes)


class AccountPageFormsHook(ExtensionHook, metaclass=ExtensionHookPoint):
    """A hook for adding new forms to a page in the My Account page.

    This is used to add custom forms to a page in the My Account page. The
    form can be used to provide user-level customization of an extension,
    through a traditional form-based approach or even through custom
    JavaScript.

    This hook takes the ID of a registered page where the form should be
    placed. Review Board supplies the following built-in page IDs:

    * ``settings``
    * ``authentication``
    * ``profile``
    * ``groups``

    Any registered page ID can be provided, whether from this extension
    or another.

    Form classes can only be added to a single page.
    """

    def initialize(self, page_id, form_classes):
        """Initialize the hook.

        This will register each of the provided page form classes on the
        account page matching the provided ID.

        Args:
            page_id (unicode):
                The page ID corresponding to a registered
                :py:class:`~reviewboard.accounts.pages.AccountPage`.

            form_classes (list of type):
                The list of form classes to register on the page. Each class
                must be a subclass of
                :py:class:`~reviewboard.accounts.forms.pages.AccountPageForm`.
        """
        self.page_id = page_id
        self.form_classes = form_classes

        page_class = AccountPage.registry.get('page_id', page_id)
        assert page_class is not None

        for form_class in form_classes:
            page_class.add_form(form_class)

    def shutdown(self):
        """Shut down the hook.

        This will unregister each of the page form classes from the associated
        page.
        """
        page_class = AccountPage.registry.get('page_id', self.page_id)
        assert page_class is not None

        for form_class in self.form_classes:
            page_class.remove_form(form_class)



"""Change WebAPIToken to allow for different token formats.

This increases the max length for tokens and adds a ``token_generator_id``
field to allow for new token formats other than the existing SHA1 format.
This also adds fields to support the expiration and invalidation of tokens.

Version Added:
    5.0
"""

from django_evolution.mutations import AddField, ChangeField
from django.db import models


MUTATIONS = [
    ChangeField('WebAPIToken', 'token', max_length=255),
    AddField('WebAPIToken', 'token_generator_id', models.CharField,
             max_length=255, initial='legacy_sha1'),
    AddField('WebAPIToken', 'last_used', models.DateTimeField, null=True),
    AddField('WebAPIToken', 'expires', models.DateTimeField, null=True),
    AddField('WebAPIToken', 'valid', models.BooleanField, initial=True),
    AddField('WebAPIToken', 'invalid_date', models.DateTimeField, null=True),
    AddField('WebAPIToken', 'invalid_reason', models.TextField, initial=''),
]



"""A hook for adding capabilities to the API server info payload."""

from __future__ import annotations

from djblets.extensions.hooks import ExtensionHook, ExtensionHookPoint

from reviewboard.webapi.server_info import (register_webapi_capabilities,
                                            unregister_webapi_capabilities)


class WebAPICapabilitiesHook(ExtensionHook, metaclass=ExtensionHookPoint):
    """This hook allows adding capabilities to the web API server info.

    Note that this does not add the functionality, but adds to the server
    info listing.

    Extensions may only provide one instance of this hook. All capabilities
    must be registered at once.
    """

    def initialize(self, caps):
        """Initialize the hook.

        This will register each of the capabilities for the API.

        Args:
            caps (dict):
                The dictionary of capabilities to register. Each key msut
                be a string, and each value should be a boolean or a
                dictionary of string keys to booleans.

        Raises:
            KeyError:
                Capabilities have already been registered by this extension.
        """
        register_webapi_capabilities(self.extension.id, caps)

    def shutdown(self):
        """Shut down the hook.

        This will unregister each of the capabilities from the API.
        """
        unregister_webapi_capabilities(self.extension.id)



"""Context processors for theme-related state.

Version Added:
    7.0
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from django.utils.html import format_html_join

from reviewboard.themes.ui.registry import ui_theme_registry

if TYPE_CHECKING:
    from django.http import HttpRequest


def theme(
    request: HttpRequest,
) -> dict[str, object]:
    """Return template context variables for theme information.

    Version Added:
        7.0

    Args:
        request (django.http.HttpRequest):
            The HTTP request from the client.

    Returns:
        dict:
        The context variables for the page.
    """
    if request.user.is_authenticated:
        profile = request.user.get_profile()
        ui_theme_id = profile.ui_theme_id
    else:
        profile = None
        ui_theme_id = 'default'

    ui_theme = (ui_theme_registry.get_theme(ui_theme_id) or
                ui_theme_registry.get_theme('default'))
    assert ui_theme is not None

    return {
        'ui_theme_attrs_html': format_html_join(
            ' ',
            '{}="{}"',
            ui_theme.get_html_attrs(request=request).items()),
    }



"""Signal handlers for review and review request related objects.

Version Added:
    6.0
"""

from __future__ import annotations

from django.db.models.signals import pre_delete

from reviewboard.reviews.models import (ReviewRequest,
                                        ReviewRequestDraft)
from reviewboard.reviews.models.review_request import FileAttachmentState


def _on_review_request_draft_deleted(
    sender: type[ReviewRequestDraft],
    instance: ReviewRequestDraft,
    using: str,
    **kwargs,
) -> None:
    """Handle any extra cleanup when a review request draft is deleted.

    Version Added:
        6.0

    Args:
        sender (type, unused):
            The sender of the signal.

        instance (reviewboard.reviews.models.ReviewRequestDraft):
            The review request draft being deleted.

        using (str, unused):
            The database alias being used.

        **kwargs (dict, unused):
            Unused additional keyword arguments.
    """
    draft_attachments = instance.get_file_attachments()

    if draft_attachments:
        # Load the data for the file attachments states.
        instance.get_file_attachments_data(
            draft_active_attachments=draft_attachments)

    for attachment in draft_attachments:
        state = instance.get_file_attachment_state(attachment)

        if state in (FileAttachmentState.NEW_REVISION,
                     FileAttachmentState.NEW):
            # This never has been and never will be published, so delete it.
            attachment.delete()

    changedesc = instance.changedesc

    if changedesc and not changedesc.review_request.exists():
        changedesc.delete()

    diffset = instance.diffset

    if (diffset and
        diffset.pk is not None and
        diffset.history is None):
        diffset.delete()

    review_request = instance.get_review_request()
    review_request.clear_local_caches()


def _on_review_request_deleted(
    sender: type[ReviewRequest],
    instance: ReviewRequest,
    using: str,
    **kwargs,
) -> None:
    """Handle extra cleanup when a review request is deleted.

    Version Added:
        7.0

    Args:
        sender (type, unused):
            The sender of the signal.

        instance (reviewboard.reviews.models.ReviewRequest):
            The review request being deleted.

        using (str, unused):
            The database alias being used.

        **kwargs (dict, unused):
            Unused additional keyword arguments.
    """
    # We defined a bunch of these things using ManyToMany fields, when they
    # really ought to have been foreign keys. We therefore can't rely on
    # cascade, and need to clean them all up manually when the review request
    # is deleted.
    instance.changedescs.all().delete()

    if (instance.file_attachments_count or
        instance.inactive_file_attachments_count):
        instance.file_attachment_histories.all().delete()

        if instance.file_attachments_count:
            instance.file_attachments.all().delete()

        if instance.inactive_file_attachments_count:
            instance.inactive_file_attachments.all().delete()

    if instance.screenshots_count:
        instance.screenshots.all().delete()

    if instance.inactive_screenshots_count:
        instance.inactive_screenshots.all().delete()

    instance.diffset_history.delete()


def connect_signal_handlers() -> None:
    """Connect review and review request related signal handlers.

    Version Added:
        6.0
    """
    pre_delete.connect(_on_review_request_draft_deleted,
                       sender=ReviewRequestDraft)
    pre_delete.connect(_on_review_request_deleted,
                       sender=ReviewRequest)



"""Built-in actions for the admins app.

Version Added:
    8.0
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from django.contrib.auth.models import User
from django.core.cache import cache
from django.db.models import Model, QuerySet
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from djblets.cache.backend import cache_memoize, make_cache_key

from reviewboard.actions import (ActionPlacement,
                                 AttachmentPoint,
                                 BaseAction,
                                 BaseGroupAction,
                                 actions_registry)
from reviewboard.actions.base import ActionAttachmentPoint
from reviewboard.actions.renderers import (SidebarActionGroupRenderer,
                                           SidebarItemActionRenderer)
from reviewboard.hostingsvcs.models import HostingServiceAccount
from reviewboard.notifications.models import WebHookTarget
from reviewboard.oauth.models import Application
from reviewboard.reviews.models import DefaultReviewer, Group
from reviewboard.scmtools.models import Repository

if TYPE_CHECKING:
    from collections.abc import Iterator
    from typing import Any, ClassVar, Final

    from django.http.request import HttpRequest
    from django.template import Context
    from typelets.django.strings import StrOrPromise


logger = logging.getLogger(__name__)


class BaseAdminSidebarGroupAction(BaseGroupAction):
    """Base class for sidebar group actions.

    This allows for additional metadata for template hooks that render
    at the end of the items for the group.

    Version Added:
        8.0
    """

    #: A template hook point to include at the bottom of the items.
    sidebar_hook_point: ClassVar[str | None] = None


class BaseAdminSidebarManageItemAction(BaseAction):
    """Base class for sidebar item actions.

    This allows for additional metadata for optional Add Item URLs/titles
    and item counts.

    Version Added:
        8.0
    """

    #: The admin site name used for URL resolution.
    #:
    #: Extensions should set this to their extension's main module name
    #: containing the admin site.
    admin_site_name: ClassVar[str] = 'admin'

    #: URL name for an Add Item icon.
    add_item_url_name: ClassVar[str | None] = None

    #: Title text used for an Add Item icon.
    add_item_title: ClassVar[StrOrPromise | None] = None

    #: Queryset used to count items in the database.
    #:
    #: If not set, a default queryset based on :py:attr:`model` will be used.
    item_queryset: ClassVar[QuerySet | None] = None

    #: The model represented by this sidebar item.
    model: ClassVar[type[Model] | None] = None

    def get_add_item_url(self) -> str | None:
        """Return the URL to add a new item.

        If :py:attr:`add_item_url_name` is set, it will be used as a URL
        name and resolved. Otherwise, a URL will be derived from
        :py:attr:`model` using the default database.

        Extensions must override to generate a path suitable for their own
        database.

        Returns:
            str:
            The Add Item URL, or ``None`` if one is not set or can't be looked
            up.
        """
        if (url_name := self.add_item_url_name):
            return reverse(url_name)
        elif self.model is not None:
            return reverse(self._get_admin_url_name('add'))

        return None

    def get_url(
        self,
        *,
        context: Context,
    ) -> str:
        """Return the URL for the navigation item.

        If :py:attr:`url` or :py:attr:`url_name` are not explicitly set,
        a URL will be generated based on :py:attr:`model`.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            str:
            The URL to use for the action.
        """
        url = super().get_url(context=context)

        if url == '#' and self.model is not None:
            url = reverse(self._get_admin_url_name('changelist'))

        return url

    def get_item_count(self) -> int | None:
        """Return the number of items to show in the sidebar.

        This will default to performing a query using
        :py:attr:`item_queryset`, returning ``None`` if not set.

        Returns:
            int:
            The number of items, or ``None`` if it can't be queried.
        """
        if (item_queryset := self.get_item_queryset()) is not None:
            return item_queryset.count()

        return None

    def get_item_queryset(self) -> QuerySet | None:
        """Return the queryset for the items.

        This will default to returning a copy of :py:attr:`item_queryset`,
        if set, or a queryset based on :py:attr:`model`.

        Returns:
            django.db.models.query.QuerySet:
            The queryset used to perform item lookups.
        """
        if (queryset := self.item_queryset) is not None:
            # Create a local copy of the queryset defined on the class.
            return queryset.all()
        elif (model := self.model) is not None:
            return model.objects.all()

        return None

    def _get_admin_url_name(
        self,
        page_name: str,
    ) -> str:
        """Return a URL name for an admin page.

        This will generate the name based on :py:attr:`admin_site_name`,
        :py:attr:`model`, and the provided page name within the admin site.

        Args:
            page_name (str):
                The name of the page to link to.

        Returns:
            str:
            The resulting admin URL name.
        """
        model = self.model
        assert model is not None

        meta = model._meta

        return (
            f'{self.admin_site_name}:{meta.app_label}_{meta.model_name}_'
            f'{page_name}'
        )


class AdminSidebarActionGroupRenderer(SidebarActionGroupRenderer):
    """Renderer for admin sidebar groups.

    This provides additional CSS for the group based on whether there are
    icons shown in group items.

    It also provides a template hook for post-group rendering, if the
    action group has a ``sidebar_hook_point`` attribute.

    Version Added:
        8.0
    """

    template_name = 'admin/actions/sidebar_group_action.html'


#
# "Administration" nav group actions.
#

class AdminMainNavGroupAction(BaseAdminSidebarGroupAction):
    """Administration navigation group action.

    Version Added:
        8.0
    """

    action_id = 'admin-main-nav-group'
    label = _('Administration')
    sidebar_hook_point = 'admin-sidebar-administration'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.ADMIN_NAV),
    ]


class AdminDashboardNavAction(BaseAction):
    """Administration -> Dashboard navigation action.

    Version Added:
        8.0
    """

    action_id = 'admin-dashboard-nav'
    label = _('Dashboard')
    url_name = 'admin-dashboard'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.ADMIN_NAV,
                        parent_id=AdminMainNavGroupAction.action_id),
    ]


class AdminLicensesNavAction(BaseAction):
    """Administration -> Licenses navigation action.

    Version Added:
        8.0
    """

    action_id = 'admin-licenses-nav'
    label = _('Licenses')
    url_name = 'admin-licenses'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.ADMIN_NAV,
                        parent_id=AdminMainNavGroupAction.action_id),
    ]


class AdminSecurityCenterNavAction(BaseAction):
    """Administration -> Security navigation action.

    Version Added:
        8.0
    """

    action_id = 'admin-security-center-nav'
    label = _('Security Center')
    url_name = 'admin-security-checks'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.ADMIN_NAV,
                        parent_id=AdminMainNavGroupAction.action_id),
    ]


class AdminExtensionsNavAction(BaseAction):
    """Administration -> Extensions navigation action.

    Version Added:
        8.0
    """

    action_id = 'admin-extensions-nav'
    label = _('Extensions')
    url_name = 'extension-list'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.ADMIN_NAV,
                        parent_id=AdminMainNavGroupAction.action_id),
    ]


class AdminIntegrationsNavAction(BaseAction):
    """Administration -> Integrations navigation action.

    Version Added:
        8.0
    """

    action_id = 'admin-integrations-nav'
    label = _('Integrations')
    url_name = 'integration-list'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.ADMIN_NAV,
                        parent_id=AdminMainNavGroupAction.action_id),
    ]


class AdminDatabaseNavAction(BaseAction):
    """Administration -> Database navigation action.

    Version Added:
        8.0
    """

    action_id = 'admin-database-nav'
    label = _('Database')
    url_name = 'admin:index'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.ADMIN_NAV,
                        parent_id=AdminMainNavGroupAction.action_id),
    ]


#
# "Settings" nav group actions.
#

class AdminSettingsNavGroupAction(BaseAdminSidebarGroupAction):
    """Settings navigation group action.

    Version Added:
        8.0
    """

    action_id = 'admin-settings-nav-group'
    label = _('Settings')
    sidebar_hook_point = 'admin-sidebar-settings'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.ADMIN_NAV),
    ]


class AdminSettingsGeneralNavAction(BaseAction):
    """Settings -> General navigation action.

    Version Added:
        8.0
    """

    action_id = 'admin-settings-general-nav'
    label = _('General')
    url_name = 'settings-general'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.ADMIN_NAV,
                        parent_id=AdminSettingsNavGroupAction.action_id),
    ]


class AdminSettingsAuthenticationNavAction(BaseAction):
    """Settings -> Authentication navigation action.

    Version Added:
        8.0
    """

    action_id = 'admin-settings-auth-nav'
    label = _('Authentication')
    url_name = 'settings-authentication'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.ADMIN_NAV,
                        parent_id=AdminSettingsNavGroupAction.action_id),
    ]


class AdminSettingsAvatarsNavAction(BaseAction):
    """Settings -> Avatars navigation action.

    Version Added:
        8.0
    """

    action_id = 'admin-settings-avatars-nav'
    label = _('Avatars')
    url_name = 'settings-avatars'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.ADMIN_NAV,
                        parent_id=AdminSettingsNavGroupAction.action_id),
    ]


class AdminSettingsEmailNavAction(BaseAction):
    """Settings -> E-Mail navigation action.

    Version Added:
        8.0
    """

    action_id = 'admin-settings-email-nav'
    label = _('E-Mail')
    url_name = 'settings-email'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.ADMIN_NAV,
                        parent_id=AdminSettingsNavGroupAction.action_id),
    ]


class AdminSettingsReviewsNavAction(BaseAction):
    """Settings -> Review Workflow navigation action.

    Version Added:
        8.0
    """

    action_id = 'admin-settings-reviews-nav'
    label = _('Review Workflow')
    url_name = 'settings-reviews'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.ADMIN_NAV,
                        parent_id=AdminSettingsNavGroupAction.action_id),
    ]


class AdminSettingsDiffsNavAction(BaseAction):
    """Settings -> Diff Viewer navigation action.

    Version Added:
        8.0
    """

    action_id = 'admin-settings-diffs-nav'
    label = _('Diff Viewer')
    url_name = 'settings-diffs'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.ADMIN_NAV,
                        parent_id=AdminSettingsNavGroupAction.action_id),
    ]


class AdminSettingsLoggingNavAction(BaseAction):
    """Settings -> Logging navigation action.

    Version Added:
        8.0
    """

    action_id = 'admin-settings-logging-nav'
    label = _('Logging')
    url_name = 'settings-logging'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.ADMIN_NAV,
                        parent_id=AdminSettingsNavGroupAction.action_id),
    ]


class AdminSettingsSSHNavAction(BaseAction):
    """Settings -> SSH navigation action.

    Version Added:
        8.0
    """

    action_id = 'admin-settings-ssh-nav'
    label = _('SSH')
    url_name = 'settings-ssh'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.ADMIN_NAV,
                        parent_id=AdminSettingsNavGroupAction.action_id),
    ]


class AdminSettingsStorageNavAction(BaseAction):
    """Settings -> File Storage navigation action.

    Version Added:
        8.0
    """

    action_id = 'admin-settings-storage-nav'
    label = _('File Storage')
    url_name = 'settings-storage'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.ADMIN_NAV,
                        parent_id=AdminSettingsNavGroupAction.action_id),
    ]


class AdminSettingsPrivacyNavAction(BaseAction):
    """Settings -> User Privacy navigation action.

    Version Added:
        8.0
    """

    action_id = 'admin-settings-privacy-nav'
    label = _('User Privacy')
    url_name = 'settings-privacy'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.ADMIN_NAV,
                        parent_id=AdminSettingsNavGroupAction.action_id),
    ]


class AdminSettingsSupportNavAction(BaseAction):
    """Settings -> Support navigation action.

    Version Added:
        8.0
    """

    action_id = 'admin-settings-support-nav'
    label = _('Support')
    url_name = 'settings-support'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.ADMIN_NAV,
                        parent_id=AdminSettingsNavGroupAction.action_id),
    ]


class AdminSettingsSearchNavAction(BaseAction):
    """Settings -> Search navigation action.

    Version Added:
        8.0
    """

    action_id = 'admin-settings-search-nav'
    label = _('Search')
    url_name = 'settings-search'

    placements = [
        ActionPlacement(attachment=AttachmentPoint.ADMIN_NAV,
                        parent_id=AdminSettingsNavGroupAction.action_id),
    ]


#
# "Manage" nav group actions.
#

class AdminSidebarManageActionRenderer(SidebarItemActionRenderer):
    """Admin sidebar renderer for items in the Manage section.

    This renders the sidebar item with an Add Item icon and an item
    count, if found. Data is pulled from cache, and recomputed if missing
    from cache.

    Version Added:
        8.0
    """

    template_name = 'admin/actions/sidebar_item_action.html'

    def get_extra_context(
        self,
        *,
        request: HttpRequest,
        context: Context,
    ) -> dict[str, Any]:
        """Return extra template context for the action.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            context (django.template.Context):
                The current rendering context.

        Returns:
            dict:
            Extra context to use when rendering the action's template.
        """
        extra_context = super().get_extra_context(request=request,
                                                  context=context)

        action = self.action
        action_id = action.action_id

        add_item_url: (str | None) = None
        add_item_title: (str | None) = None
        item_count: (int | None) = None

        # Determine the Add Item URL to show.
        try:
            add_item_url = context['add_item_urls'][action_id]
        except KeyError:
            add_item_url = _get_manage_action_item_url(action)

        # Determine the item count.
        try:
            item_count = context['item_counts'][action_id]
        except KeyError:
            item_count = _get_manage_action_item_count(action)

        if add_item_url:
            # Determine a title.
            add_item_title = getattr(action, 'add_item_title', None)

            if (not add_item_title and
                (model := getattr(action, 'model', None)) is not None):
                # Compute a default title for the model.
                add_item_title = (
                    _('Add a new {model}')
                    .format(model=model._meta.verbose_name)
                )

        extra_context.update({
            'add_item_url': add_item_url,
            'add_item_title': add_item_title,
            'item_count': item_count,
        })

        return extra_context


class AdminSidebarManageActionGroupRenderer(AdminSidebarActionGroupRenderer):
    """Admin sidebar renderer for the Manage section.

    This is a special action group that looks up and caches both the
    Add Item URLs for each type of action and the number of items currently
    in the database.

    State is cached especially to reduce the number of database queries
    across page loads. Caches are only valid for 5 minutes, ensuring that
    manual modifications to the database won't retain old counts for long.

    Cached state is opportunistic. URLs and counts for any given item are
    fetched by the item renderer as needed if not found in the cache.

    Version Added:
        8.0
    """

    default_item_renderer_cls = AdminSidebarManageActionRenderer

    #: The cache key used to store item counts.
    _STATE_CACHE_KEY: Final[str] = 'admin-sidebar-manage-state'

    def get_extra_context(
        self,
        *,
        request: HttpRequest,
        context: Context,
    ) -> dict[str, Any]:
        """Return extra template context for the action.

        This includes precomputed state for our default actions, containing
        Add Item URLs and item counts.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            context (django.template.Context):
                The current rendering context.

        Returns:
            dict:
            Extra context to use when rendering the action's template.
        """
        extra_context = super().get_extra_context(request=request,
                                                  context=context)

        # Cache item counts to avoid repeated database queries during
        # navigation. This cache will be invalidated when tracked objects
        # are created or deleted.
        sidebar_state = cache_memoize(self._STATE_CACHE_KEY,
                                      self._get_sidebar_state)

        extra_context.update(sidebar_state)

        return extra_context

    def _get_sidebar_state(self) -> dict[str, Any]:
        """Return state for items in the sidebar.

        This iterates through all the items within the Manage section of
        the admin sidebar and computed URLs and counts for each, returning
        dictionary for the context.

        The result can be cached for future page lookups.

        Returns:
            dict:
            Context state for the cache.
        """
        # We're precomputing URLs in here, rather than computing them in the
        # template, because we need to always ensure that reverse() will be
        # searching all available URL patterns and not just the ones bound to
        # request.current_app.
        #
        # current_app gets set by AdminSite views, and if we're in an
        # extension's AdminSite view, we'll fail to resolve these URLs from
        # within the template. We don't have that problem if calling reverse()
        # ourselves.
        add_item_urls: dict[str, str] = {}
        item_counts: dict[str, int] = {}
        models: set[type[Model]] = set()

        for action in self.placement.child_actions:
            action_id = action.action_id

            if (url := _get_manage_action_item_url(action)) is not None:
                add_item_urls[action_id] = url

            if (count := _get_manage_action_item_count(action)) is not None:
                item_counts[action_id] = count

            if (model := getattr(action, 'model', None)) is not None:
                models.add(model)

        setattr(self.action, '_managed_models', models)

        return {
            'add_item_urls': add_item_urls,
            'item_counts': item_counts,
        }


class AdminManageNavGroupAction(BaseAdminSidebarGroupAction):
    """Manage navigation group action.

    Version Added:
        8.0
    """

    action_id = 'admin-manage-nav-group'
    label = _('Manage')
    sidebar_hook_point = 'admin-sidebar-manage'

    default_renderer_cls = AdminSidebarManageActionGroupRenderer

    placements = [
        ActionPlacement(attachment=AttachmentPoint.ADMIN_NAV),
    ]


class AdminManageUsersNavAction(BaseAdminSidebarManageItemAction):
    """Manage -> Users navigation action.

    Version Added:
        8.0
    """

    action_id = 'admin-manage-users-nav'
    label = _('Users')
    model = User

    placements = [
        ActionPlacement(attachment=AttachmentPoint.ADMIN_NAV,
                        parent_id=AdminManageNavGroupAction.action_id),
    ]


class AdminManageReviewGroupsNavAction(BaseAdminSidebarManageItemAction):
    """Manage -> Review Groups navigation action.

    Version Added:
        8.0
    """

    action_id = 'admin-manage-review-groups-nav'
    label = _('Review Groups')
    model = Group

    placements = [
        ActionPlacement(attachment=AttachmentPoint.ADMIN_NAV,
                        parent_id=AdminManageNavGroupAction.action_id),
    ]


class AdminManageDefaultReviewersNavAction(BaseAdminSidebarManageItemAction):
    """Manage -> Default Reviewers navigation action.

    Version Added:
        8.0
    """

    action_id = 'admin-manage-default-reviewers-nav'
    label = _('Default Reviewers')
    model = DefaultReviewer

    placements = [
        ActionPlacement(attachment=AttachmentPoint.ADMIN_NAV,
                        parent_id=AdminManageNavGroupAction.action_id),
    ]


class AdminManageRepositoriesNavAction(BaseAdminSidebarManageItemAction):
    """Manage -> Repositories navigation action.

    Version Added:
        8.0
    """

    action_id = 'admin-manage-repositories-nav'
    label = _('Repositories')
    model = Repository

    placements = [
        ActionPlacement(attachment=AttachmentPoint.ADMIN_NAV,
                        parent_id=AdminManageNavGroupAction.action_id),
    ]


class AdminManageWebHooksNavAction(BaseAdminSidebarManageItemAction):
    """Manage -> WebHooks navigation action.

    Version Added:
        8.0
    """

    action_id = 'admin-manage-webhooks-nav'
    label = _('WebHooks')
    model = WebHookTarget

    placements = [
        ActionPlacement(attachment=AttachmentPoint.ADMIN_NAV,
                        parent_id=AdminManageNavGroupAction.action_id),
    ]


class AdminManageHostingAccountsNavAction(BaseAdminSidebarManageItemAction):
    """Manage -> Hosting Accounts navigation action.

    Version Added:
        8.0
    """

    action_id = 'admin-manage-hosting-accounts-nav'
    label = _('Hosting Accounts')
    model = HostingServiceAccount

    placements = [
        ActionPlacement(attachment=AttachmentPoint.ADMIN_NAV,
                        parent_id=AdminManageNavGroupAction.action_id),
    ]


class AdminManageOAuth2AppsNavAction(BaseAdminSidebarManageItemAction):
    """Manage -> OAuth2 Applications navigation action.

    Version Added:
        8.0
    """

    action_id = 'admin-manage-oauth-app-nav'
    label = _('OAuth2 Applications')
    model = Application

    placements = [
        ActionPlacement(attachment=AttachmentPoint.ADMIN_NAV,
                        parent_id=AdminManageNavGroupAction.action_id),
    ]


def get_default_admin_actions() -> Iterator[BaseAction]:
    """Return the default actions registered for the admin UI.

    This supplies default data for the actions registry. It's not meant to
    be called by any other callers.

    Version Added:
        8.0

    Yields:
        reviewboard.actions.base.BaseAction:
        Each action to register.
    """
    yield from (
        # Administration section
        AdminMainNavGroupAction(),
        AdminDashboardNavAction(),
        AdminLicensesNavAction(),
        AdminSecurityCenterNavAction(),
        AdminExtensionsNavAction(),
        AdminIntegrationsNavAction(),
        AdminDatabaseNavAction(),

        # Settings section
        AdminSettingsNavGroupAction(),
        AdminSettingsGeneralNavAction(),
        AdminSettingsAuthenticationNavAction(),
        AdminSettingsAvatarsNavAction(),
        AdminSettingsEmailNavAction(),
        AdminSettingsReviewsNavAction(),
        AdminSettingsDiffsNavAction(),
        AdminSettingsLoggingNavAction(),
        AdminSettingsSSHNavAction(),
        AdminSettingsStorageNavAction(),
        AdminSettingsPrivacyNavAction(),
        AdminSettingsSupportNavAction(),
        AdminSettingsSearchNavAction(),

        # Manage section
        AdminManageNavGroupAction(),
        AdminManageUsersNavAction(),
        AdminManageReviewGroupsNavAction(),
        AdminManageDefaultReviewersNavAction(),
        AdminManageRepositoriesNavAction(),
        AdminManageWebHooksNavAction(),
        AdminManageHostingAccountsNavAction(),
        AdminManageOAuth2AppsNavAction(),
    )


def get_default_admin_attachment_points() -> Iterator[ActionAttachmentPoint]:
    """Return the default attachment points registered for the admin UI.

    This supplies default data for the attachment points registry. It's not
    meant to be called by any other callers.

    Version Added:
        8.0

    Yields:
        reviewboard.actions.base.ActionAttachmentPoint:
        Each attachment point to register.
    """
    yield ActionAttachmentPoint(
        AttachmentPoint.ADMIN_NAV,
        default_action_renderer_cls=SidebarItemActionRenderer,
        default_action_group_renderer_cls=AdminSidebarActionGroupRenderer,
    )


def _get_manage_action_item_url(
    action: BaseAction,
) -> str | None:
    """Return the URL to add a new item for an action.

    This will only return a URL for actions subclassing
    :py:class:`BaseAdminSidebarManageItemAction`.

    Any errors will be caught and logged.

    Version Added:
        8.0

    Args:
        action (reviewboard.actions.base.BaseAction):
            The action for which to return a URL.

    Returns:
        str:
        The Add Item URL, or ``None`` if one is not set or can't be looked
        up.
    """
    if isinstance(action, BaseAdminSidebarManageItemAction):
        try:
            return action.get_add_item_url()
        except Exception as e:
            logger.exception('Unexpected error getting Add Item URL for '
                             'action "%s": %s',
                             action.action_id, e)

    return None


def _get_manage_action_item_count(
    action: BaseAction,
) -> int | None:
    """Return the number of items backed by an action.

    This will only return a count for actions subclassing
    :py:class:`BaseAdminSidebarManageItemAction`.

    Any errors will be caught and logged.

    Version Added:
        8.0

    Args:
        action (reviewboard.actions.base.BaseAction):
            The action for which to return a count.

    Returns:
        str:
        The number of items, or ``None`` if a count can't be returned.
    """
    if isinstance(action, BaseAdminSidebarManageItemAction):
        try:
            return action.get_item_count()
        except Exception as e:
            logger.exception('Unexpected error querying item count for '
                             'action "%s": %s',
                             action.action_id, e)

    return None


def _invalidate_cache_for_model(
    model: type[Model],
) -> None:
    """Invalidate caches used for the sidebar state based on a model change.

    If the model is managed by the sidebar, the cache state will be
    invalidated.

    Version Added:
        8.0

    Args:
        model (type):
            The type of model required to invalidate the cache.
    """
    action = actions_registry.get_action(AdminManageNavGroupAction.action_id)

    if (action is not None and
        model in getattr(action, '_managed_models', [])):
        # This model's represented in the sidebar. The cache can be
        # invalidated.
        cache.delete(make_cache_key(
            AdminSidebarManageActionGroupRenderer._STATE_CACHE_KEY))


@receiver(post_save)
def _on_post_save(
    *,
    sender: type[Model],
    created: bool = False,
    **kwargs,
) -> None:
    """Clear cached action state when a model is saved.

    Whenever any model is created in the database, the cached state for the
    admin sidebar will be invalidated, preventing stale counts from being
    shown.

    Version Added:
        8.0

    Args:
        sender (type):
            The type of model that was saved.

        created (bool, optional):
            Whether the model was created.

        **kwargs (tuple, unused):
            Keyword arguments passed to the signal handler.
    """
    if created:
        _invalidate_cache_for_model(sender)


@receiver(post_delete)
def _on_post_delete(
    *,
    sender: type[Model],
    **kwargs,
) -> None:
    """Clear cached action state when a model is deleted.

    Whenever any model is deleted in the database, the cached state for the
    admin sidebar will be invalidated, preventing stale counts from being
    shown.

    Version Added:
        8.0

    Args:
        sender (type):
            The type of model that was saved.

        **kwargs (tuple, unused):
            Keyword arguments passed to the signal handler.
    """
    _invalidate_cache_for_model(sender)



"""Unit tests for SAML forms."""

from __future__ import annotations

from typing import ClassVar, TYPE_CHECKING

from django.contrib.auth.models import User
from djblets.siteconfig.models import SiteConfiguration

from reviewboard.accounts.sso.backends import sso_backends
from reviewboard.accounts.sso.backends.saml.forms import (SAMLLinkUserForm,
                                                          SAMLSettingsForm)
from reviewboard.accounts.sso.backends.saml.settings import (
    SAMLBinding,
    SAMLDigestAlgorithm,
    SAMLNameIDFormat,
    SAMLSignatureAlgorithm)
from reviewboard.testing import TestCase

if TYPE_CHECKING:
    from typelets.json import JSONDict


VALID_CERT = """-----BEGIN CERTIFICATE-----
MIICZjCCAc+gAwIBAgIBADANBgkqhkiG9w0BAQ0FADBQMQswCQYDVQQGEwJ1czEL
MAkGA1UECAwCQ0ExFjAUBgNVBAoMDUJlYW5iYWcsIEluYy4xHDAaBgNVBAMME2h0
dHBzOi8vZXhhbXBsZS5jb20wHhcNMjIwNTA2MTU0NjI1WhcNMjMwNTA2MTU0NjI1
WjBQMQswCQYDVQQGEwJ1czELMAkGA1UECAwCQ0ExFjAUBgNVBAoMDUJlYW5iYWcs
IEluYy4xHDAaBgNVBAMME2h0dHBzOi8vZXhhbXBsZS5jb20wgZ8wDQYJKoZIhvcN
AQEBBQADgY0AMIGJAoGBANCsbj4mvUiQERBy80R7yqA6hU3FMM4siC2UcUS3ltFF
grkVOAPr+zUnrdadmAiTH35AB94oMzf0Qh8OJCr7wG5JQm686TRkVm2xUxhJUcoq
7LjBTKeEXBcrEzdNlagFXxHUSz5bPSdwDt/zbOfe+9RZKeb4FggFCEYw/mi69+Dx
AgMBAAGjUDBOMB0GA1UdDgQWBBS4cP9Y+IM7ZHZChUDdx68QExTZUDAfBgNVHSME
GDAWgBS4cP9Y+IM7ZHZChUDdx68QExTZUDAMBgNVHRMEBTADAQH/MA0GCSqGSIb3
DQEBDQUAA4GBALht5/NfJU+GxYfQKiGkZ4Ih/T/48rzXAT7/7f61s7w72UR2S5e2
WsR7/JPkZ5+u5mCgmABjNcd9NzaBM2RfSrrurwbjXMQ8nb/+REvhXXJ4STsS48y5
bef2JtIf7mGDw8/KsUrAA2jEIpCedToGyQxyE6GdN5b69ITWvyAemnIM
-----END CERTIFICATE-----"""


class SAMLLinkUserFormTests(TestCase):
    """Unit tests for SAMLLinkUserForm."""

    fixtures: ClassVar[list[str]] = ['test_users']

    def test_valid_login(self) -> None:
        """Testing SAMLLinkUserForm validation in login mode"""
        form = SAMLLinkUserForm(data={
            'username': 'doc',
            'password': 'doc',
            'provision': False,
        })

        self.assertTrue(form.is_valid())

    def test_invalid_login(self) -> None:
        """Testing SAMLLinkUserForm validation with incorrect password in login
        mode
        """
        form = SAMLLinkUserForm(data={
            'username': 'doc',
            'password': 'nope',
            'provision': False,
        })

        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors['__all__'],
                         ['Please enter a correct username and password. Note '
                          'that both fields may be case-sensitive.'])

    def test_inactive_user(self) -> None:
        """Testing SAMLLinkUserForm validation with an inactive user"""
        user = User.objects.get(username='doc')
        user.is_active = False
        user.save(update_fields=['is_active'])

        form = SAMLLinkUserForm(data={
            'username': 'doc',
            'password': 'nope',
            'provision': False,
        })

        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors['__all__'],
                         ['This account is inactive.'])

    def test_provision(self) -> None:
        """Testing SAMLLinkUserForm validation in provision mode"""
        form = SAMLLinkUserForm(data={
            'username': 'doc',
            'password': 'nope',
            'provision': True,
        })

        self.assertTrue(form.is_valid())

    def test_provision_disabled(self) -> None:
        """Testing SAMLLinkUserForm validation in provision mode"""
        form = SAMLLinkUserForm(data={
            'username': 'doc',
            'password': 'nope',
            'provision': True,
        })

        settings: JSONDict = {
            'saml_automatically_provision_users': False,
        }

        with self.siteconfig_settings(settings):
            self.assertFalse(form.is_valid())
            self.assertEqual(
                form.errors['__all__'],
                ['A user account for doc does not exist. Your administrator '
                 'will need to provision an account before you can log in.'])


class SAMLSettingsFormTests(TestCase):
    """Unit tests for SAMLSettingsForm."""

    def setUp(self):
        """Set up the test case."""
        self.siteconfig = SiteConfiguration.objects.get_current()

        saml_backend = sso_backends.get('backend_id', 'saml')

        # Ensure everything is set to defaults.
        for key, value in saml_backend.siteconfig_defaults.items():
            self.siteconfig.set(key, value)

        self.siteconfig.save()

    def test_save(self):
        """Testing SAMLSettingsForm.save"""
        siteconfig = self.siteconfig

        form = SAMLSettingsForm(
            siteconfig,
            data={
                'saml_login_button_text': 'Login',
                'saml_issuer': 'https://example.com/saml/issuer',
                'saml_signature_algorithm': SAMLSignatureAlgorithm.RSA_SHA1,
                'saml_digest_algorithm': SAMLDigestAlgorithm.SHA512,
                'saml_nameid_format': SAMLNameIDFormat.PERSISTENT,
                'saml_verification_cert': VALID_CERT,
                'saml_sso_url': 'https://example.com/saml/sso',
                'saml_sso_binding_type': SAMLBinding.HTTP_POST,
                'saml_slo_url': 'https://example.com/saml/slo',
                'saml_slo_binding_type': SAMLBinding.HTTP_REDIRECT,
                'saml_require_login_to_link': False,
                'saml_attr_email': 'User.Email',
                'saml_attr_firstname': 'User.FirstName',
                'saml_attr_lastname': 'User.LastName',
                'saml_attr_fullname': 'User.FullName',
            })

        self.assertTrue(form.is_valid())
        form.save()

        siteconfig.refresh_from_db()
        self.assertEqual(siteconfig.get('saml_login_button_text'),
                         'Login')
        self.assertEqual(siteconfig.get('saml_issuer'),
                         'https://example.com/saml/issuer')
        self.assertEqual(siteconfig.get('saml_signature_algorithm'),
                         SAMLSignatureAlgorithm.RSA_SHA1)
        self.assertEqual(siteconfig.get('saml_digest_algorithm'),
                         SAMLDigestAlgorithm.SHA512)
        self.assertEqual(siteconfig.get('saml_nameid_format'),
                         SAMLNameIDFormat.PERSISTENT)
        self.assertEqual(siteconfig.get('saml_verification_cert'),
                         VALID_CERT)
        self.assertEqual(siteconfig.get('saml_sso_url'),
                         'https://example.com/saml/sso')
        self.assertEqual(siteconfig.get('saml_sso_binding_type'),
                         SAMLBinding.HTTP_POST)
        self.assertEqual(siteconfig.get('saml_slo_url'),
                         'https://example.com/saml/slo')
        self.assertEqual(siteconfig.get('saml_slo_binding_type'),
                         SAMLBinding.HTTP_REDIRECT)
        self.assertFalse(siteconfig.get('saml_require_login_to_link'))



"""Diff parsing support."""

from __future__ import annotations

import io
import logging
import re
import weakref
from copy import deepcopy
from typing import Optional, Sequence, TYPE_CHECKING, Union

from django.utils.translation import gettext as _
from djblets.util.properties import TypedProperty
from housekeeping import deprecate_non_keyword_only_args
from pydiffx import DiffType, DiffX
from pydiffx.errors import DiffXParseError
from typing_extensions import TypeAlias

from reviewboard.deprecation import RemovedInReviewBoard10_0Warning
from reviewboard.diffviewer.errors import DiffParserError
from reviewboard.scmtools.core import HEAD, PRE_CREATION, Revision, UNKNOWN

if TYPE_CHECKING:
    from pydiffx import BaseDiffXSection
    from typelets.json import JSONDict

    from reviewboard.diffviewer.models import DiffCommit
    from reviewboard.diffviewer.models import DiffSet


logger = logging.getLogger(__name__)


#: An alias for a TypedProperty taking a bytes value or None.
#:
#: Version Added:
#:     6.0
_BytesProperty: TypeAlias = TypedProperty[Optional[bytes], Optional[bytes]]

#: An alias for a TypedProperty taking a bytes value, Revision, or None.
#:
#: Version Added:
#:     6.0
_RevisionProperty: TypeAlias = TypedProperty[
    Optional[Union[bytes, Revision]],
    Optional[Union[bytes, Revision]]
]

#: An alias for a TypedProperty taking a str value or None.
#:
#: Version Added:
#:     6.0
_StrProperty: TypeAlias = TypedProperty[Optional[str], Optional[str]]


class ParsedDiff:
    """Parsed information from a diff.

    This stores information on the diff as a whole, along with a list of
    commits made to the diff and a list of files within each.

    Extra data can be stored by the parser, which will be made available in
    :py:attr:`DiffSet.extra_data
    <reviewboard.diffviewer.models.diffset.DiffSet.extra_data>`.

    This is flexible enough to accommodate a variety of diff formats,
    including DiffX files.

    This class is meant to be used internally and by subclasses of
    :py:class:`BaseDiffParser`.

    Version Added:
        4.0.5
    """

    ######################
    # Instance variables #
    ######################

    #: The list of changes parsed in this diff.
    #:
    #: There should always be at least one.
    changes: list[ParsedDiffChange]

    #: Extra data to store along with the information on the diff.
    #:
    #: The contents will be stored directly in :py:attr:`DiffSet.extra_data
    #: <reviewboard.diffviewer.models.diffset.DiffSet.extra_data>`.
    extra_data: JSONDict

    #: The diff parser that parsed this file.
    parser: BaseDiffParser

    #: Whether commit IDs are used as file revisions.
    #:
    #: A commit ID will be used if an explicit revision isn't available for
    #: a file. For instance, if a parent diff is available, and a file isn't
    #: present in the parent diff, the file will use the parent diff's
    #: parent commit ID as the parent revision.
    uses_commit_ids_as_revisions: bool

    @deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
    def __init__(
        self,
        *,
        parser: BaseDiffParser,
        uses_commit_ids_as_revisions: bool = False,
        has_per_file_revisions: bool = True,
    ) -> None:
        """Initialize the parsed diff information.

        Version Changed:
            7.0.2:
            * All arguments must now be passed as keyword arguments.
            * Added the ``has_per_file_revisions`` argument.

        Args:
            parser (BaseDiffParser):
                The diff parser that parsed this file.

            uses_commit_ids_as_revisions (bool, optional):
                Whether commit IDs are used as file revisions.

                See :py:attr:`ParsedDiff.uses_commit_ids_as_revisions`.

            has_per_file_revisions (bool, optional):
                Whether the diff has revisions listed per-file.

                Most SCM diffs list a source revision for each file that can
                be directly tied to the change that introduced the file, but
                some do not.

                If set to ``False``, :py:attr:`extra_data` will be updated
                with a ``has_per_file_revisions=False`` flag set, helping
                consumers of the diff handle file lookup in an alternate
                way.

                Version Added:
                    7.0.2
        """
        self.parser = parser
        self.extra_data = {}
        self.changes = []
        self.uses_commit_ids_as_revisions = uses_commit_ids_as_revisions

        if not has_per_file_revisions:
            self.extra_data['has_per_file_revisions'] = False


class ParsedDiffChange:
    """Parsed change information from a diff.

    This stores information on a change to a tree, consisting of a set of
    parsed files and extra data to store (in :py:attr:`DiffCommit.extra_data
    <reviewboard.diffviewer.models.diffcommit.DiffCommit.extra_data>`.

    This will often map to a commit, or just a typical collection of files in a
    diff. Traditional diffs will have only one of these. DiffX files may have
    many (but for the moment, only diffs with a single change can be handled
    when processing these results).

    Version Added:
        4.0.5
    """

    #: The ID of the commit, parsed from the diff.
    #:
    #: This may be ``None``.
    #:
    #: Type:
    #:     bytes
    commit_id: _BytesProperty = TypedProperty(bytes)

    #: The ID of the primary parent commit, parsed from the diff.
    #:
    #: This may be ``None``.
    #:
    #: Type:
    #:     bytes
    parent_commit_id: _BytesProperty = TypedProperty(bytes)

    ######################
    # Instance variables #
    ######################

    #: Extra data to store along with the information on the change.
    #:
    #: The contents will be stored directly in :py:attr:`DiffCommit.extra_data
    #: <reviewboard.diffviewer.models.diffcommit.DiffCommit.extra_data>`.
    extra_data: JSONDict

    #: The list of files parsed for this change.
    #:
    #: There should always be at least one.
    files: list[ParsedDiffFile]

    #: A weak reference to the parent ParsedDiff.
    _parent: weakref.ref[ParsedDiff]

    @deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
    def __init__(
        self,
        *,
        parsed_diff: ParsedDiff,
    ) -> None:
        """Initialize the parsed diff information.

        Version Changed:
            7.0.2:
            All arguments must now be passed as keyword arguments.

        Args:
            parsed_diff (ParsedDiff):
                The parent parsed diff information.
        """
        assert parsed_diff is not None

        self._parent = weakref.ref(parsed_diff)
        self.extra_data = {}
        self.files = []

        parsed_diff.changes.append(self)

    @property
    def parent_parsed_diff(self) -> Optional[ParsedDiff]:
        """The parent diff object.

        Type:
            ParsedDiff
        """
        if self._parent:
            return self._parent()

        return None


class ParsedDiffFile:
    """A parsed file from a diff.

    This stores information on a single file represented in a diff, including
    the contents of that file's diff, as parsed by :py:class:`DiffParser` or
    one of its subclasses.

    Parsers should set the attributes on this based on the contents of the
    diff, and should add any data found in the diff.

    This class is meant to be used internally and by subclasses of
    :py:class:`BaseDiffParser`.

    Version Changed:
        4.0.6:
        Added :py:attr:`old_symlink_target` and py:attr:`new_symlink_target`.

    Version Changed:
        4.0.5:
        Diff parsers that manually construct instances must pass in
        ``parsed_diff_change`` instead of ``parser`` when constructing the
        object, and must call :py:meth:`discard` after construction if the
        file isn't wanted in the results.
    """

    #: The parsed original name of the file.
    #:
    #: Type:
    #:     bytes
    orig_filename: _BytesProperty = TypedProperty(bytes)

    #: The parsed file details of the original file.
    #:
    #: This will usually be a revision.
    #:
    #: Type:
    #:     bytes or reviewboard.scmtools.core.Revision
    orig_file_details: _RevisionProperty = TypedProperty((bytes, Revision))

    #: The parsed modified name of the file.
    #:
    #: This may be the same as :py:attr:`orig_filename`.
    #:
    #: Type:
    #:     bytes
    modified_filename: _BytesProperty = TypedProperty(bytes)

    #: The parsed file details of the modified file.
    #:
    #: This will usually be a revision.
    #:
    #: Type:
    #:     bytes or reviewboard.scmtools.core.Revision
    modified_file_details: _RevisionProperty = TypedProperty((bytes,
                                                              Revision))

    #: The parsed value for an Index header.
    #:
    #: If present in the diff, this usually contains a filename, but may
    #: contain other content as well, depending on the variation of the diff
    #: format.
    #:
    #: Type:
    #:     bytes
    index_header_value: _BytesProperty = TypedProperty(bytes)

    #: The old target for a symlink.
    #:
    #: Version Added:
    #:     4.0.6
    #:
    #: Type:
    #:     bytes
    old_symlink_target: _BytesProperty = TypedProperty(bytes)

    #: The new target for a symlink.
    #:
    #: Version Added:
    #:     4.0.6
    #:
    #: Type:
    #:     bytes
    new_symlink_target: _BytesProperty = TypedProperty(bytes)

    #: The old UNIX mode for the file.
    #:
    #: Version Added:
    #:     4.0.6
    #:
    #: Type:
    #:     str
    old_unix_mode: _StrProperty = TypedProperty(str)

    #: The new UNIX mode for the file.
    #:
    #: Version Added:
    #:     4.0.6
    #:
    #: Type:
    #:     str
    new_unix_mode: _StrProperty = TypedProperty(str)

    ######################
    # Instance variables #
    ######################

    #: Whether this represents a binary file.
    binary: bool

    #: Whether this represents a file that has been copied.
    #:
    #: The file may or may not be modified in the process.
    copied: bool

    #: Whether this represents a file that has been deleted.
    deleted: bool

    #: The number of delete (``-``) lines found in the file.
    delete_count: int

    #: Extra data to store along with the information on the file.
    #:
    #: The contents will be stored directly in :py:attr:`FileDiff.extra_data
    #: <reviewboard.diffviewer.models.filediff.FileDiff.extra_data>`.
    extra_data: JSONDict

    #: The number of insert (``+``) lines found in the file.
    insert_count: int

    #: Whether this represents a file that is a symbolic link to another file.
    is_symlink: bool

    #: Whether this represents a file that has been moved/renamed.
    #:
    #: The file may or may not be modified in the process.
    moved: bool

    #: The diff parser that parsed this file.
    parser: BaseDiffParser

    #: Whether this file should be skipped by the parser.
    #:
    #: If any of the parser methods set this, the file will stop parsing
    #: and will be excluded from results.
    skip: bool

    #: The finalized diff data.
    _data: Optional[bytes]

    #: The buffer of data to write before finalizing.
    _data_io: io.BytesIO

    #: A weak reference to the parent ParsedDiffChange.
    _parent: Optional[weakref.ref[ParsedDiffChange]]

    @deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
    def __init__(
        self,
        *,
        parser: Optional[BaseDiffParser] = None,
        parsed_diff_change: Optional[ParsedDiffChange] = None,
        **kwargs,
    ) -> None:
        """Initialize the parsed file information.

        Version Changed:
            7.0.2:
            * The ``parser`` deprecation and ``parsed_diff_change`` requirement
              have been re-scheduled for Review Board 8.
            * All arguments must now be passed as keyword arguments.

        Version Changed:
            4.0.5:
            Added the ``parsed_diff_change`` argument (which will be required
            in Review Board 6.0).

            Deprecated the ``parser`` argument (which will be removed in
            Review Board 6.0).

        Args:
            parser (reviewboard.diffviewer.parser.BaseDiffParser, optional):
                The diff parser that parsed this file.

                This is deprecated and will be remoed in Review Board 6.0.

            parsed_diff_change (ParsedDiffChange, optional):
                The diff change that owns this file.

                This will be required in Review Board 6.0.
        """
        if parser is not None or parsed_diff_change is None:
            RemovedInReviewBoard10_0Warning.warn(
                'ParsedDiffFile.__init__(parser=[...]) is deprecated in favor '
                'of parsed_diff_change=[...], which will be required in '
                'Review Board 10.')

        if parsed_diff_change is not None:
            parsed_diff_change.files.append(self)

            parent_parsed_diff = parsed_diff_change.parent_parsed_diff
            assert parent_parsed_diff is not None

            parser = parent_parsed_diff.parser

            parsed_diff_change_ref = weakref.ref(parsed_diff_change)
        else:
            parsed_diff_change_ref = None

        assert parser is not None

        self._parent = parsed_diff_change_ref
        self.parser = parser
        self.binary = False
        self.deleted = False
        self.moved = False
        self.copied = False
        self.is_symlink = False
        self.insert_count = 0
        self.delete_count = 0
        self.skip = False
        self.extra_data = {}

        self._data_io = io.BytesIO()
        self._data = None

    @property
    def parent_parsed_diff_change(self) -> Optional[ParsedDiffChange]:
        """The parent change object.

        Version Added:
            4.0.5

        Type:
            ParsedDiffChange
        """
        if self._parent:
            return self._parent()

        return None

    @property
    def data(self) -> bytes:
        """The data for this diff.

        This must be accessed after :py:meth:`finalize` has been called.
        """
        if self._data is None:
            raise ValueError('ParsedDiffFile.data cannot be accessed until '
                             'finalize() is called.')

        return self._data

    def discard(self) -> None:
        """Discard this from the parent change.

        This will remove it from the list of files. It's intended for use
        when a diff parser is populating the diff but then determines the
        file is no longer needed.

        Version Added:
            4.0.5
        """
        assert self.parent_parsed_diff_change

        self.parent_parsed_diff_change.files.remove(self)

    def finalize(self) -> None:
        """Finalize the parsed diff.

        This makes the diff data available to consumers and closes the buffer
        for writing.
        """
        self._data = self._data_io.getvalue()
        self._data_io.close()

    def prepend_data(
        self,
        data: bytes,
    ) -> None:
        """Prepend data to the buffer.

        Args:
            data (bytes):
                The data to prepend.
        """
        if data:
            new_data_io = io.BytesIO()
            new_data_io.write(data)
            new_data_io.write(self._data_io.getvalue())

            self._data_io.close()
            self._data_io = new_data_io

    def append_data(
        self,
        data: bytes,
    ) -> None:
        """Append data to the buffer.

        Args:
            data (bytes):
                The data to append.
        """
        if data:
            self._data_io.write(data)


class BaseDiffParser:
    """Base class for a diff parser.

    This is a low-level, basic foundational interface for a diff parser. It
    performs type checking of the incoming data and a couple of methods for
    subclasses to implement.

    Most SCM implementations will want to either subclass
    :py:class:`DiffParser` or use :py:class:`DiffXParser`.

    Version Added:
        4.0.5
    """

    ######################
    # Instance variables #
    ######################

    #: The diff data being parsed.
    data: bytes

    #: Whether commit IDs are used as file revisions.
    #:
    #: See :py:attr:`ParsedDiff.uses_commit_ids_as_revisions`.
    uses_commit_ids_as_revisions: bool

    #: Whether the diff has revisions listed per-file.
    #:
    #: Most SCM diffs list a source revision for each file that can be
    #: directly tied to the change that introduced the file, but some
    #: (Mercurial) do not.
    #:
    #: Diff parsers that don't support per-file revisions must set this to
    #: ``False`` so that consumers can alter their file lookup logic.
    #:
    #: Version Added:
    #:     7.0.2
    has_per_file_revisions: bool = True

    @deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
    def __init__(
        self,
        *,
        data: bytes,
        uses_commit_ids_as_revisions: bool = False,
        **kwargs,
    ) -> None:
        """Initialize the parser.

        Version Changed:
            7.0.2:
            All arguments must now be passed as keyword arguments.

        Args:
            data (bytes):
                The diff content to parse.

            uses_commit_ids_as_revisions (bool):
                Whether commit IDs are used as file revisions.

                See :py:attr:`ParsedDiff.uses_commit_ids_as_revisions`.

            **kwargs (dict, unused):
                Unused keyword arguments.

        Raises:
            TypeError:
                The provided ``data`` argument was not a ``bytes`` type.
        """
        if not isinstance(data, bytes):
            raise TypeError(
                _('%s expects bytes values for "data", not %s')
                % (type(self).__name__, type(data)))

        self.data = data
        self.uses_commit_ids_as_revisions = uses_commit_ids_as_revisions

    def parse_diff(self) -> ParsedDiff:
        """Parse the diff.

        This will parse the content of the file, returning a representation
        of the diff file and its content.

        This must be implemented by subclasses.

        Returns:
            ParsedDiff:
            The resulting parsed diff information.

        Raises:
            NotImplementedError:
                This wasn't implemented by a subclass.

            reviewboard.diffviewer.errors.DiffParserError:
                There was an error parsing part of the diff. This may be a
                corrupted diff, or an error in the parsing implementation.
                Details are in the error message.
        """
        raise NotImplementedError

    def raw_diff(
        self,
        diffset_or_commit: Union[DiffCommit, DiffSet],
    ) -> bytes:
        """Return a raw diff as a string.

        This takes a DiffSet or DiffCommit and generates a new, single diff
        file that represents all the changes made. It's used to regenerate
        a diff and serve it up for other tools or processes to use.

        This must be implemented by subclasses.

        Args:
            diffset_or_commit (reviewboard.diffviewer.models.diffset.DiffSet or
                               reviewboard.diffviewer.models.diffcommit
                               .DiffCommit):
                The DiffSet or DiffCommit to render.

                If passing in a DiffSet, only the cumulative diff's file
                contents will be returned.

                If passing in a DiffCommit, only that commit's file contents
                will be returned.

        Returns:
            bytes:
            The diff composed of all the component FileDiffs.

        Raises:
            NotImplementedError:
                This wasn't implemented by a subclass.

            TypeError:
                The provided ``diffset_or_commit`` wasn't of a supported type.
        """
        raise NotImplementedError

    def normalize_diff_filename(
        self,
        filename: str,
    ) -> str:
        """Normalize filenames in diffs.

        This returns a normalized filename suitable for populating in
        :py:attr:`FileDiff.source_file
        <reviewboard.diffviewer.models.filediff.FileDiff.source_file>` or
        :py:attr:`FileDiff.dest_file
        <reviewboard.diffviewer.models.filediff.FileDiff.dest_file>`, or
        for when presenting a filename to the UI.

        By default, this strips off any leading slashes, which might occur due
        to differences in various diffing methods or APIs.

        Subclasses can override this to provide additional methods of
        normalization.

        Args:
            filename (str):
                The filename to normalize.

        Returns:
            str:
            The normalized filename.
        """
        if filename.startswith('/'):
            return filename[1:]
        else:
            return filename


class DiffParser(BaseDiffParser):
    """Parses diff files, allowing subclasses to specialize parsing behavior.

    This class provides the base functionality for parsing Unified Diff files.
    It looks for common information present in many variations of diffs,
    such as ``Index:`` lines, in order to extract files and their modified
    content from a diff.

    Subclasses can extend the parsing behavior to extract additional metadata
    or handle special representations of changes. They may want to override the
    following methods:

    * :py:meth:`parse_special_header`
    * :py:meth:`parse_diff_header`
    * :py:meth:`parse_filename_header`
    * :py:meth:`parse_after_headers`
    * :py:meth:`normalize_diff_filename`
    """

    #: A separator string below an Index header.
    #:
    #: This is commonly found immediately below an ``Index:`` header, meant
    #: to help locate the beginning of the metadata or changes made to a file.
    #:
    #: Its presence and location is not guaranteed.
    INDEX_SEP = b'=' * 67

    ######################
    # Instance variables #
    ######################

    #: The ID of the commit this change is based on.
    base_commit_id: Optional[str]

    #: The list of resuting parsed diff files from this parser.
    files: list[ParsedDiffFile]

    #: The diff content, split into lines.
    lines: list[bytes]

    #: The new commit ID, if available.
    new_commit_id: Optional[str]

    #: The parsed diff object.
    parsed_diff: ParsedDiff

    #: The parsed diff change object.
    parsed_diff_change: ParsedDiffChange

    def __init__(
        self,
        data: bytes,
        **kwargs,
    ) -> None:
        """Initialize the parser.

        Version Changed:
            4.0.5:
            Added ``**kwargs``.

        Args:
            data (bytes):
                The diff content to parse.

            **kwargs (dict):
                Keyword arguments to pass to the parent class.

                Version Added:
                    4.0.5

        Raises:
            TypeError:
                The provided ``data`` argument was not a ``bytes`` type.
        """
        from reviewboard.diffviewer.diffutils import split_line_endings

        super().__init__(data=data,
                         **kwargs)

        self.base_commit_id = None
        self.new_commit_id = None
        self.lines = split_line_endings(data)

        self.parsed_diff = ParsedDiff(
            parser=self,
            uses_commit_ids_as_revisions=self.uses_commit_ids_as_revisions,
            has_per_file_revisions=self.has_per_file_revisions)
        self.parsed_diff_change = ParsedDiffChange(
            parsed_diff=self.parsed_diff)

    def parse_diff(self) -> ParsedDiff:
        """Parse the diff.

        Subclasses should override this if working with a diff format that
        extracts more than one change from a diff.

        Version Added:
            4.0.5:
            Historically, :py:meth:`parse` was the main method used to parse a
            diff. That's now used exclusively to parse a list of files for
            the default :py:attr:`parsed_diff_change`. The old method is
            around for compatibility, but is no longer called directly outside
            of this class.

        Returns:
            ParsedDiff:
            The resulting parsed diff information.

        Raises:
            reviewboard.diffviewer.errors.DiffParserError:
                There was an error parsing part of the diff. This may be a
                corrupted diff, or an error in the parsing implementation.
                Details are in the error message.
        """
        class_name = type(self).__name__
        logger.debug('%s.parse_diff: Beginning parse of diff, size = %s',
                     class_name, len(self.data))

        parsed_diff_files = self.parse()
        parsed_diff_change = self.parsed_diff_change

        for parsed_diff_file in parsed_diff_files:
            if parsed_diff_file.parent_parsed_diff_change is None:
                parsed_diff_change.files.append(parsed_diff_file)

        logger.debug('%s.parse_diff: Finished parsing diff.', class_name)

        return self.parsed_diff

    def parse(self) -> Sequence[ParsedDiffFile]:
        """Parse the diff and return a list of files.

        This will parse the content of the file, returning any files that
        were found.

        Version Change:
            4.0.5:
            Historically, this was the main method used to parse a diff. It's
            now used exclusively to parse a list of files for the default
            :py:attr:`parsed_diff_change`, and :py:meth:`parse_diff` is the
            main method used to parse a diff. This method is around for
            compatibility, but is no longer called directly outside of this
            class.

        Returns:
            list of ParsedDiffFile:
            The resulting list of files.

        Raises:
            reviewboard.diffviewer.errors.DiffParserError:
                There was an error parsing part of the diff. This may be a
                corrupted diff, or an error in the parsing implementation.
                Details are in the error message.
        """
        preamble: io.BytesIO = io.BytesIO()
        self.files = []
        parsed_file: Optional[ParsedDiffFile] = None
        i: int = 0

        # Go through each line in the diff, looking for diff headers.
        while i < len(self.lines):
            next_linenum, new_file = self.parse_change_header(i)

            if new_file:
                # This line is the start of a new file diff.
                #
                # First, finalize the last one.
                if self.files:
                    self.files[-1].finalize()

                parsed_file = new_file
                assert parsed_file is not None

                # We need to prepend the preamble, if we have one.
                parsed_file.prepend_data(preamble.getvalue())

                preamble.close()
                preamble = io.BytesIO()

                self.files.append(parsed_file)
                i = next_linenum
            else:
                if parsed_file:
                    i = self.parse_diff_line(i, parsed_file)
                else:
                    preamble.write(self.lines[i])
                    preamble.write(b'\n')
                    i += 1

        if self.files:
            self.files[-1].finalize()

        preamble.close()

        return self.files

    def parse_diff_line(
        self,
        linenum: int,
        parsed_file: ParsedDiffFile,
    ) -> int:
        """Parse a line of data in a diff.

        This will append the line to the parsed file's data, and if the
        content represents active changes to a file, its insert/delete counts
        will be updated to reflect them.

        Args:
            linenum (int):
                The 0-based line number.

            parsed_file (ParsedDiffFile):
                The current parsed diff file info.

        Returns:
            int:
            The next line number to parse.
        """
        line = self.lines[linenum]

        if (parsed_file.orig_filename is not None and
            parsed_file.modified_filename is not None):
            if line.startswith(b'-'):
                parsed_file.delete_count += 1
            elif line.startswith(b'+'):
                parsed_file.insert_count += 1

        parsed_file.append_data(line)
        parsed_file.append_data(b'\n')

        return linenum + 1

    def parse_change_header(
        self,
        linenum: int,
    ) -> tuple[int, Optional[ParsedDiffFile]]:
        """Parse a header before a change to a file.

        This will attempt to parse the following information, starting at the
        specified line in the diff:

        1. Any special file headers (such as ``Index:`` lines) through
           :py:meth:`parse_special_header`
        2. A standard Unified Diff file header (through
           :py:meth:`parse_diff_header`)
        3. Any content after the header (through
           :py:meth:`parse_after_headers`)

        If the special or diff headers are able to populate the original and
        modified filenames and revisions/file details, and none of the methods
        above mark the file as skipped (by setting
        :py:attr:`ParsedDiffFile.skip`), then this will finish by appending
        all parsed data and returning a parsed file entry.

        Subclasses that need to control parsing logic should override one or
        more of the above methods.

        Args:
            linenum (int):
                The line number to begin parsing.

        Returns:
            tuple:
            A tuple containing the following:

            Tuple:
                0 (int):
                    The next line number to parse.

                1 (ParsedDiffFile):
                    The populated :py:class:`ParsedDiffFile` instance for
                    this file

        Raises:
            reviewboard.diffviewer.errors.DiffParserError:
                There was an error parsing the change header. This may be
                a corrupted diff, or an error in the parsing implementation.
                Details are in the error message.
        """
        parsed_file: Optional[ParsedDiffFile] = \
            ParsedDiffFile(parsed_diff_change=self.parsed_diff_change)
        assert parsed_file is not None

        start = linenum

        linenum = self.parse_special_header(linenum, parsed_file)
        linenum = self.parse_diff_header(linenum, parsed_file)

        skip = (
            parsed_file.skip or
            parsed_file.orig_filename is None or
            parsed_file.orig_file_details is None or
            parsed_file.modified_filename is None or
            parsed_file.modified_file_details is None
        )

        if not skip:
            # If we have enough information to represent a header, build the
            # file to return.
            if linenum < len(self.lines):
                linenum = self.parse_after_headers(linenum, parsed_file)

                skip = parsed_file.skip

        if skip:
            parsed_file.discard()
            parsed_file = None
        else:
            # The header is part of the diff, so make sure it gets in the
            # diff content.
            for line in self.lines[start:linenum]:
                parsed_file.append_data(line)
                parsed_file.append_data(b'\n')

        return linenum, parsed_file

    def parse_special_header(
        self,
        linenum: int,
        parsed_file: ParsedDiffFile,
    ) -> int:
        """Parse a special diff header marking the start of a new file's info.

        This attempts to locate an ``Index:`` line at the specified line
        number, which usually indicates the beginning of file's information in
        a diff (for Unified Diff variants that support it). By default, this
        method expects the line to be found at ``linenum``.

        If present, the value found immediately after the ``Index:`` will be
        stored in :py:attr:`ParsedDiffFile.index_header_value`, allowing
        subclasses to make a determination based on its contents (which may
        vary between types of diffs, but should include at least a filename.

        If the ``Index:`` line is not present, this won't do anything by
        default.

        Subclasses can override this to parse additional information before the
        standard diff header. They may also set :py:attr:`ParsedFileDiff.skip`
        to skip the rest of this file and begin parsing a new entry at the
        returned line number.

        Args:
            linenum (int):
                The line number to begin parsing.

            parsed_file (ParsedDiffFile):
                The file currently being parsed.

        Returns:
            int:
            The next line number to parse.

        Raises:
            reviewboard.diffviewer.errors.DiffParserError:
                There was an error parsing the special header. This may be
                a corrupted diff, or an error in the parsing implementation.
                Details are in the error message.
        """
        try:
            index_line = self.lines[linenum]
            is_index = index_line.startswith(b'Index: ')
        except IndexError:
            index_line = None
            is_index = False

        if is_index:
            assert index_line is not None

            # Try to find the "====" line.
            temp_linenum = linenum + 1

            while temp_linenum + 1 < len(self.lines):
                line = self.lines[temp_linenum]

                if line == self.INDEX_SEP:
                    # We found the line. This is looking like a valid diff
                    # for CVS, Subversion, and other systems. Try to parse
                    # the data from the line.
                    try:
                        parsed_file.index_header_value = \
                            index_line.split(None, 1)[1]
                    except ValueError:
                        raise DiffParserError('Malformed Index line', linenum)

                    linenum = temp_linenum + 1
                    break
                elif line.startswith((b'---', b'+++')):
                    # We never found that line, but we did hit the start of
                    # a diff file. We can't treat the "Index:" line as special
                    # in this case.
                    break

                temp_linenum += 1

        return linenum

    def parse_diff_header(
        self,
        linenum: int,
        parsed_file: ParsedDiffFile,
    ) -> int:
        """Parse a standard header before changes made to a file.

        This attempts to parse the ``---`` (original) and ``+++`` (modified)
        file lines, which are usually present right before any changes to the
        file. By default, this method expects the ``---`` line to be found at
        ``linenum``.

        If found, this will populate :py:attr:`ParsedDiffFile.orig_filename`,
        :py:attr:`ParsedDiffFile.orig_file_details`,
        :py:attr:`ParsedDiffFile.modified_filename`, and
        :py:attr:`ParsedDiffFile.modified_file_details`.

        This calls out to :py:meth:`parse_filename_header` to help parse
        the contents immediately after the ``---`` or ``+++``.

        Subclasses can override this to parse these lines differently, or to
        to process the results of these lines (such as converting special
        filenames to states like "deleted" or "new file"). They may also set
        :py:class:`ParsedFileDiff.skip` to skip the rest of this file and begin
        parsing a new entry at the returned line number.

        Args:
            linenum (int):
                The line number to begin parsing.

            parsed_file (ParsedDiffFile):
                The file currently being parsed.

        Returns:
            int:
            The next line number to parse.

        Raises:
            reviewboard.diffviewer.errors.DiffParserError:
                There was an error parsing the diff header. This may be a
                corrupted diff, or an error in the parsing implementation.
                Details are in the error message.
        """
        try:
            line1 = self.lines[linenum]
            line2 = self.lines[linenum + 1]

            is_diff_header = (
                # Unified diff headers
                (line1.startswith(b'--- ') and line2.startswith(b'+++ ')) or

                # Context diff headers
                (line1.startswith(b'*** ') and line2.startswith(b'--- ') and
                 not line1.endswith(b' ****'))
            )
        except IndexError:
            is_diff_header = False

        if is_diff_header:
            # This is a unified or context diff header. Parse the
            # file and extra info.
            try:
                (parsed_file.orig_filename,
                 parsed_file.orig_file_details) = \
                    self.parse_filename_header(self.lines[linenum][4:],
                                               linenum)
                linenum += 1

                (parsed_file.modified_filename,
                 parsed_file.modified_file_details) = \
                    self.parse_filename_header(self.lines[linenum][4:],
                                               linenum)

                linenum += 1
            except ValueError:
                raise DiffParserError(
                    'The diff file is missing revision information',
                    linenum)

        return linenum

    def parse_after_headers(
        self,
        linenum: int,
        parsed_file: ParsedDiffFile,
    ) -> int:
        """Parse information after a diff header but before diff data.

        This attempts to parse the information found after
        :py:meth:`parse_diff_headers` is called, but before gathering any lines
        that are part of the diff contents. It's intended for the few diff
        formats that may place content at this location.

        By default, this does nothing.

        Subclasses can override this to provide custom parsing of any lines
        that may exist here. They may also set :py:class:`ParsedFileDiff.skip`
        to skip the rest of this file and begin parsing a new entry at the
        returned line number.

        Args:
            linenum (int):
                The line number to begin parsing.

            parsed_file (ParsedDiffFile):
                The file currently being parsed.

        Returns:
            int:
            The next line number to parse.

        Raises:
            reviewboard.diffviewer.errors.DiffParserError:
                There was an error parsing the diff header. This may be a
                corrupted diff, or an error in the parsing implementation.
                Details are in the error message.
        """
        return linenum

    def parse_filename_header(
        self,
        s: bytes,
        linenum: int,
    ) -> tuple[bytes, bytes]:
        """Parse the filename found in a diff filename line.

        This parses the value after a ``---`` or ``+++`` indicator (or a
        special variant handled by a subclass), normalizing the filename and
        any following file details, and returning both for processing and
        storage.

        Often times, the file details will be a revision for the original
        file, but this is not guaranteed, and is up to the variation of the
        diff format.

        By default, this will assume that a filename and file details are
        separated by either a single tab, or two or more spaces. If neither
        are found, this will fail to parse.

        This must parse only the provided value, and cannot parse subsequent
        lines.

        Subclasses can override this behavior to parse these lines another
        way, or to normalize filenames (handling escaping or filenames with
        spaces as needed by that particular diff variation).

        Args:
            s (bytes):
                The value to parse.

            linenum (int):
                The line number containing the value to parse.

        Returns:
            tuple:
            A tuple containing:

            Tuple:
                0 (bytes):
                    The filename.

                1 (bytes):
                    The additional file information (as bytes).

        Raises:
            reviewboard.diffviewer.errors.DiffParserError:
                There was an error parsing the diff header. This may be a
                corrupted diff, or an error in the parsing implementation.
                Details are in the error message.
        """
        if b'\t' in s:
            # There's a \t separating the filename and info. This is the
            # best case scenario, since it allows for filenames with spaces
            # without much work.
            filename, info = s.split(b'\t', 1)

            return filename, info

        # There's spaces being used to separate the filename and info.
        # This is technically wrong, so all we can do is assume that
        # 1) the filename won't have multiple consecutive spaces, and
        # 2) there's at least 2 spaces separating the filename and info.
        if b'  ' in s:
            filename, info = re.split(br'  +', s, 1)

            return filename, info

        raise DiffParserError('No valid separator after the filename was '
                              'found in the diff header',
                              linenum)

    def raw_diff(
        self,
        diffset_or_commit: Union[DiffCommit, DiffSet],
    ) -> bytes:
        """Return a raw diff as a string.

        This takes a DiffSet or DiffCommit and generates a new, single diff
        file that represents all the changes made. It's used to regenerate
        a diff and serve it up for other tools or processes to use.

        Subclasses can override this to provide any special logic for building
        the diff.

        Args:
            diffset_or_commit (reviewboard.diffviewer.models.diffset.DiffSet or
                               reviewboard.diffviewer.models.diffcommit
                               .DiffCommit):
                The DiffSet or DiffCommit to render.

                If passing in a DiffSet, only the cumulative diff's file
                contents will be returned.

                If passing in a DiffCommit, only that commit's file contents
                will be returned.

        Returns:
            bytes:
            The diff composed of all the component FileDiffs.

        Raises:
            TypeError:
                The provided ``diffset_or_commit`` wasn't of a supported type.
        """
        if hasattr(diffset_or_commit, 'cumulative_files'):
            # This will be a DiffSet.
            if TYPE_CHECKING:
                assert isinstance(diffset_or_commit, DiffSet)

            filediffs = diffset_or_commit.cumulative_files
        elif hasattr(diffset_or_commit, 'files'):
            # This will be a DiffCommit.
            if TYPE_CHECKING:
                assert isinstance(diffset_or_commit, DiffCommit)

            filediffs = diffset_or_commit.files.order_by('pk')
        else:
            raise TypeError('%r is not a valid value. Please pass a DiffSet '
                            'or DiffCommit.'
                            % diffset_or_commit)

        return b''.join(
            filediff.diff
            for filediff in filediffs
        )


class DiffXParser(BaseDiffParser):
    """Parser for DiffX files.

    This will parse files conforming to the DiffX_ standard, storing the
    diff content provided in each file section, as well as all the information
    available in each DiffX section (options, preamble, metadata) as
    ``extra_data``. This allows the diffs to be re-built on download.

    This parser is sufficient for most any DiffX need, but subclasses can
    be created that augment the stored ``extra_data`` for any of the parsed
    objects.

    .. _DiffX: https://diffx.org

    Version Added:
        4.0.5:
        This is experimental in 4.0.x, with plans to make it stable for 5.0.
        The API may change during this time.
    """

    def parse_diff(self) -> ParsedDiff:
        """Parse the diff.

        This will parse the content of the DiffX file, returning a
        representation of the diff file and its content.

        Returns:
            ParsedDiff:
            The resulting parsed diff information.

        Raises:
            reviewboard.diffviewer.errors.DiffParserError:
                There was an error parsing part of the diff. This may be a
                corrupted diff, or an error in the parsing implementation.
                Details are in the error message.
        """
        class_name = type(self).__name__
        logger.debug('%s.parse_diff: Beginning parse of diff, size = %s',
                     class_name, len(self.data))

        try:
            diffx = DiffX.from_bytes(self.data)
        except DiffXParseError as e:
            raise DiffParserError(str(e))

        MOVED_OPS = {
            'move',
            'move-modify',
        }

        COPIED_OPS = {
            'copy',
            'copy-modify',
        }

        # Process the main DiffX file information.
        parsed_diff = ParsedDiff(parser=self)
        parsed_diff.uses_commit_ids_as_revisions = \
            self.uses_commit_ids_as_revisions

        extra_data_diffx: JSONDict = {}
        self._store_options(extra_data_diffx, diffx)
        self._store_preamble(extra_data_diffx, diffx)
        self._store_meta(extra_data_diffx, diffx)

        if extra_data_diffx:
            parsed_diff.extra_data['diffx'] = extra_data_diffx

        # Process each change in the DiffX file.
        for change_num, diffx_change in enumerate(diffx.changes, start=1):
            parsed_diff_change = ParsedDiffChange(parsed_diff=parsed_diff)

            # Extract information and populate the ParsedDiffChange.
            change_meta = diffx_change.meta
            commit_id = change_meta.get('id')
            parent_ids = change_meta.get('parent ids')

            if commit_id is not None:
                parsed_diff_change.commit_id = commit_id.encode('utf-8')

            if parent_ids:
                parsed_diff_change.parent_commit_id = \
                    parent_ids[0].encode('utf-8')

            extra_data_change: JSONDict = {}
            self._store_options(extra_data_change, diffx_change)
            self._store_preamble(extra_data_change, diffx_change)
            self._store_meta(extra_data_change, diffx_change)

            if extra_data_change:
                parsed_diff_change.extra_data['diffx'] = extra_data_change

            # Process each file in the DiffX change.
            for file_num, diffx_file in enumerate(diffx_change.files, start=1):
                parsed_diff_file = ParsedDiffFile(
                    parsed_diff_change=parsed_diff_change)

                extra_data_file: JSONDict = {}
                self._store_options(extra_data_file, diffx_file)
                self._store_meta(extra_data_file, diffx_file)
                self._store_options(extra_data_file, diffx_file.diff_section,
                                    key='diff_options')

                if extra_data_file:
                    parsed_diff_file.extra_data['diffx'] = extra_data_file

                # Extract information needed to populate the ParsedDiffFile.
                file_meta = diffx_file.meta
                diff_data = diffx_file.diff
                path_info = file_meta.get('path')
                revision_info = file_meta.get('revision', {})
                stats_info = file_meta.get('stats')
                op = file_meta.get('op', 'modify')

                # Parse the file path information.
                if isinstance(path_info, dict):
                    # If the file is a dictionary, both keys are required.
                    try:
                        orig_filename = path_info['old']
                        modified_filename = path_info['new']
                    except KeyError as e:
                        raise DiffParserError(
                            _('Missing the "path.%(key)s" key in change '
                              '%(change_num)s, file %(file_num)s')
                            % {
                                'key': e.args[0],
                                'change_num': change_num,
                                'file_num': file_num,
                            })
                elif isinstance(path_info, str):
                    # If the file is a string, both filenames are the same.
                    orig_filename = path_info
                    modified_filename = path_info
                else:
                    raise DiffParserError(
                        _('Unexpected type %(type)s for "path" key in change '
                          '%(change_num)s, file %(file_num)s')
                        % {
                            'change_num': change_num,
                            'file_num': file_num,
                            'type': type(path_info),
                        })

                # Parse the revision information.
                if isinstance(revision_info, dict):
                    if 'old' in revision_info:
                        orig_revision = Revision(revision_info['old'])
                    else:
                        if op == 'create':
                            orig_revision = PRE_CREATION
                        else:
                            orig_revision = UNKNOWN

                    if 'new' in revision_info:
                        modified_revision = Revision(revision_info['new'])
                    else:
                        modified_revision = HEAD
                else:
                    raise DiffParserError(
                        _('Unexpected type %(type)s for "revision" key in '
                          'change %(change_num)s, file %(file_num)s')
                        % {
                            'change_num': change_num,
                            'file_num': file_num,
                            'type': type(revision_info),
                        })

                # Grab the insert/delete statistics.
                if (not stats_info or
                    'insertions' not in stats_info or
                    'deletions' not in stats_info):
                    # This DiffX is lacking stats. We'll need to generate
                    # it now.
                    #
                    # If there's a problem with the diff, then this could
                    # still fail, so we'll still need to default the values
                    # to 0 below.
                    diffx_file.generate_stats()
                    stats_info = diffx_file.meta.get('stats') or {}

                # We can now populate the ParsedDiffFile.
                parsed_diff_file.orig_filename = orig_filename.encode('utf-8')
                parsed_diff_file.orig_file_details = orig_revision
                parsed_diff_file.modified_filename = \
                    modified_filename.encode('utf-8')
                parsed_diff_file.modified_file_details = modified_revision

                parsed_diff_file.binary = \
                    (diffx_file.diff_type == DiffType.BINARY)
                parsed_diff_file.is_symlink = \
                    (file_meta.get('type') == 'symlink')

                parsed_diff_file.deleted = (op == 'delete')
                parsed_diff_file.moved = op in MOVED_OPS
                parsed_diff_file.copied = op in COPIED_OPS

                parsed_diff_file.insert_count = stats_info.get('insertions', 0)
                parsed_diff_file.delete_count = stats_info.get('deletions', 0)

                try:
                    parsed_diff_file.extra_data['encoding'] = \
                        extra_data_file['diff_options']['encoding']
                except KeyError:
                    # An explicit encoding wasn't set.
                    pass

                # If this represents a symlink, set the information.
                if parsed_diff_file.is_symlink:
                    symlink_target = file_meta.get('symlink target')

                    old_symlink_target: Optional[str]
                    new_symlink_target: Optional[str]

                    if isinstance(symlink_target, dict):
                        old_symlink_target = symlink_target.get('old')
                        new_symlink_target = symlink_target.get('new')
                    elif isinstance(symlink_target, str):
                        old_symlink_target = symlink_target
                        new_symlink_target = symlink_target
                    else:
                        logger.warning('Unexpected symlink target type (%r) '
                                       'found in diff %r',
                                       symlink_target, self.data)
                        old_symlink_target = None
                        new_symlink_target = None

                    if old_symlink_target or new_symlink_target:
                        old_symlink_target_bytes: Optional[bytes]
                        new_symlink_target_bytes: Optional[bytes]

                        if old_symlink_target:
                            old_symlink_target_bytes = \
                                old_symlink_target.encode('utf-8')
                        else:
                            old_symlink_target_bytes = None

                        if new_symlink_target:
                            new_symlink_target_bytes = \
                                new_symlink_target.encode('utf-8')
                        else:
                            new_symlink_target_bytes = None

                        if op == 'create':
                            parsed_diff_file.new_symlink_target = \
                                new_symlink_target_bytes
                        elif op == 'delete':
                            parsed_diff_file.old_symlink_target = \
                                old_symlink_target_bytes
                        else:
                            parsed_diff_file.old_symlink_target = \
                                old_symlink_target_bytes
                            parsed_diff_file.new_symlink_target = \
                                new_symlink_target_bytes

                # If there are UNIX file modes, set them.
                unix_mode = file_meta.get('unix file mode')

                if unix_mode is not None:
                    if isinstance(unix_mode, dict):
                        old_unix_mode = unix_mode.get('old')
                        new_unix_mode = unix_mode.get('new')
                    elif isinstance(unix_mode, str):
                        old_unix_mode = unix_mode
                        new_unix_mode = unix_mode
                    else:
                        logger.warning('Unexpected UNIX file mode (%r) '
                                       'found in diff %r',
                                       unix_mode, self.data)
                        old_unix_mode = None
                        new_unix_mode = None

                    if old_unix_mode or new_unix_mode:
                        if op == 'create':
                            parsed_diff_file.new_unix_mode = new_unix_mode
                        elif op == 'delete':
                            parsed_diff_file.old_unix_mode = old_unix_mode
                        else:
                            parsed_diff_file.new_unix_mode = new_unix_mode
                            parsed_diff_file.old_unix_mode = old_unix_mode

                parsed_diff_file.append_data(diff_data)
                parsed_diff_file.finalize()

        logger.debug('%s.parse_diff: Finished parsing diff.', class_name)

        return parsed_diff

    def raw_diff(
        self,
        diffset_or_commit: Union[DiffCommit, DiffSet],
    ) -> bytes:
        """Return a raw diff as a string.

        This takes a :py:class:`~reviewboard.diffviewer.models.diffset.DiffSet`
        or :py:class:`~reviewboard.diffviewer.models.diffcommit.DiffCommit` and
        generates a new, single DiffX file that represents all the changes
        made, based on the previously-stored DiffX information in
        ``extra_data`` dictionaries. It's used to regenerate a DiffX and serve
        it up for other tools or processes to use.

        Args:
            diffset_or_commit (reviewboard.diffviewer.models.diffset.DiffSet or
                               reviewboard.diffviewer.models.diffcommit
                               .DiffCommit):
                The DiffSet or DiffCommit to render.

                If passing in a DiffSet, the full uploaded DiffX file
                contents will be returned.

                If passing in a DiffCommit, a new DiffX representing only
                that commit's contents will be returned. This will lack the
                main preamble or metadata, or any other changes previously
                in the DiffX file.

        Returns:
            bytes:
            The resulting DiffX file contents.

        Raises:
            TypeError:
                The provided ``diffset_or_commit`` value wasn't of a
                supported type.
        """
        if hasattr(diffset_or_commit, 'cumulative_files'):
            # This will be a DiffSet.
            #
            # We'll pull out all the commits and files at once, to reduce
            # query counts.
            #
            # We also will be very careful about not assuming keys that are
            # present will necessarily be dictionaries. Be a bit careful and
            # default anything falsy to an empty dictionary, here and below.
            diffset = diffset_or_commit

            if TYPE_CHECKING:
                assert isinstance(diffset, DiffSet)

            diffx_main_info = diffset.extra_data.get('diffx') or {}
            diffcommits = diffset.commits.prefetch_related('files')

            if diffcommits:
                changes = [
                    {
                        'extra_data': diffcommit.extra_data,
                        'files': diffcommit.files.order_by('pk'),
                    }
                    for diffcommit in diffcommits
                ]
            else:
                changes = [
                    {
                        'extra_data': diffset.extra_data.get(
                            'change_extra_data', {}),
                        'files': diffset.cumulative_files,
                    },
                ]
        elif hasattr(diffset_or_commit, 'files'):
            # This will be a DiffCommit.
            #
            # We'll still need to pull out the DiffSet and grab the encoding,
            # if one is specified, since this will impact the DiffCommit's
            # change section.
            diffcommit = diffset_or_commit

            if TYPE_CHECKING:
                assert isinstance(diffcommit, DiffCommit)

            changes = [
                {
                    'extra_data': diffcommit.extra_data,
                    'files': diffcommit.files.order_by('pk'),
                },
            ]

            diffset_diffx_info = \
                diffcommit.diffset.extra_data.get('diffx') or {}
            diffset_diffx_options = diffset_diffx_info.get('options') or {}
            main_encoding = diffset_diffx_options.get('encoding')

            diffx_main_info: JSONDict = {}

            if main_encoding:
                diffx_main_info['options'] = {
                    'encoding': main_encoding,
                }
        else:
            raise TypeError('%r is not a valid value. Please pass a DiffSet '
                            'or DiffCommit.'
                            % diffset_or_commit)

        diffx = DiffX()
        self._load_options(diffx, diffx_main_info)
        self._load_preamble(diffx, diffx_main_info)
        self._load_meta(diffx, diffx_main_info)

        for change in changes:
            diffx_change_info = change['extra_data'].get('diffx', {})

            diffx_change = diffx.add_change()
            self._load_options(diffx_change, diffx_change_info)
            self._load_preamble(diffx_change, diffx_change_info)
            self._load_meta(diffx_change, diffx_change_info)

            for filediff in change['files']:
                diffx_file_info = filediff.extra_data.get('diffx') or {}

                diffx_file = diffx_change.add_file()
                self._load_options(diffx_file, diffx_file_info)
                self._load_meta(diffx_file, diffx_file_info)

                if filediff.diff:
                    diffx_file.diff = filediff.diff
                    self._load_options(diffx_file.diff_section,
                                       diffx_file_info,
                                       key='diff_options')

        return diffx.to_bytes()

    def _store_options(
        self,
        extra_data: JSONDict,
        diffx_section: BaseDiffXSection,
        *,
        key: str = 'options',
    ) -> None:
        """Store options for a section in extra_data.

        Options will be stored only if not empty.

        Args:
            extra_data (dict):
                The dictionary in which to store option data.

            diffx_section (pydiffx.dom.objects.BaseDiffXSection):
                The section containing the options to store.

            key (str, optional):
                The name of the key to use in ``extra_data``.
        """
        if diffx_section.options:
            extra_data[key] = deepcopy(diffx_section.options)

    def _store_preamble(
        self,
        extra_data: JSONDict,
        diffx_section: BaseDiffXSection,
    ) -> None:
        """Store preamble options and text for a section in extra_data.

        Preamble text will only be stored if not empty. Options will only
        be stored if neither is empty.

        Args:
            extra_data (dict):
                The dictionary in which to store preamble information.

            diffx_section (pydiffx.dom.objects.BaseDiffXSection):
                The section containing the preamble options and text to store.
        """
        if diffx_section.preamble:
            extra_data['preamble'] = diffx_section.preamble

            self._store_options(extra_data, diffx_section.preamble_section,
                                key='preamble_options')

    def _store_meta(
        self,
        extra_data: JSONDict,
        diffx_section: BaseDiffXSection,
    ) -> None:
        """Store metadata options and content for a section in extra_data.

        Metadata will only be stored if not empty. Options will only be
        stored if neither is empty.

        Args:
            extra_data (dict):
                The dictionary in which to store metadata information.

            diffx_section (pydiffx.dom.objects.BaseDiffXSection):
                The section containing the metadata options and content to
                store.
        """
        if diffx_section.meta:
            extra_data['metadata'] = deepcopy(diffx_section.meta)

            self._store_options(extra_data, diffx_section.meta_section,
                                key='metadata_options')

    def _load_options(
        self,
        diffx_section: BaseDiffXSection,
        extra_data: JSONDict,
        *,
        key: str = 'options',
    ) -> None:
        """Load options from extra_data into a section.

        Args:
            extra_data (dict):
                The dictionary in which to load option data.

            diffx_section (pydiffx.dom.objects.BaseDiffXSection):
                The section to store the options in.

            key (str, optional):
                The name of the key to use in ``extra_data``.
        """
        options = extra_data.get(key)

        if options:
            diffx_section.options.clear()
            diffx_section.options.update(options)

    def _load_preamble(
        self,
        diffx_section: BaseDiffXSection,
        extra_data: JSONDict,
    ) -> None:
        """Load a preamble and options from extra_data into a section.

        Args:
            extra_data (dict):
                The dictionary in which to load preamble data.

            diffx_section (pydiffx.dom.objects.BaseDiffXSection):
                The section to store the preamble in.

            key (unicode, optional):
                The name of the key to use in ``extra_data``.
        """
        preamble = extra_data.get('preamble')

        if preamble:
            diffx_section.preamble = preamble
            self._load_options(diffx_section.preamble_section,
                               extra_data,
                               key='preamble_options')

    def _load_meta(
        self,
        diffx_section: BaseDiffXSection,
        extra_data: JSONDict,
    ) -> None:
        """Load metadata and options from extra_data into a section.

        Args:
            extra_data (dict):
                The dictionary in which to load metadata information.

            diffx_section (pydiffx.dom.objects.BaseDiffXSection):
                The section to store the metadata in.

            key (unicode, optional):
                The name of the key to use in ``extra_data``.
        """
        preamble = extra_data.get('metadata')

        if preamble:
            diffx_section.meta = preamble
            self._load_options(diffx_section.meta_section,
                               extra_data,
                               key='metadata_options')



"""Template tags for actions.

Version Added:
    6.0
"""

from __future__ import annotations

import json
import logging
from typing import TYPE_CHECKING

from django.template import Context, Library
from django.utils.safestring import SafeString, mark_safe
from djblets.pagestate.state import PageState

from reviewboard.actions import (ActionAttachmentPoint,
                                 BaseAction,
                                 action_attachment_points_registry,
                                 actions_registry)
from reviewboard.actions.renderers import BaseActionGroupRenderer

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator


logger = logging.getLogger(__name__)
register = Library()


@register.simple_tag(takes_context=True)
def actions_html(
    context: Context,
    attachment: str | ActionAttachmentPoint,
) -> SafeString:
    """Render the actions HTML for the given attachment point.

    Version Changed:
        8.0:
        This now accepts an
        :py:class:`reviewboard.actions.base.ActionAttachmentPoint` instance
        for the attachment point.

    Version Added:
        6.0

    Args:
        context (django.template.Context):
            The current template rendering context.

        attachment (str or reviewboard.actions.base.ActionAttachmentPoint):
            The instance or registered ID of an attachment point.

            Version Changed:
                8.0:
                This now accepts an
                :py:class:`reviewboard.actions.base.ActionAttachmentPoint`
                instance.

    Returns:
        django.utils.safestring.SafeString:
        The rendered HTML.
    """
    request = context['request']

    # Get the referenced attachment point.
    attachment_point: ActionAttachmentPoint | None

    if isinstance(attachment, ActionAttachmentPoint):
        attachment_point = attachment
    else:
        attachment_point = (
            action_attachment_points_registry
            .get_attachment_point(attachment)
        )

        if attachment_point is None:
            logger.error('Unregistered action attachment point "%s" passed '
                         'to {% actions_html %}',
                         attachment,
                         extra={'request': request})
            return mark_safe('')

    # Store the rendered JavaScript views for later page injection.
    page_state = PageState.for_request(request)
    page_state.extra_data.setdefault('rb-actions-rendered', set())
    page_state.inject('rb-action-views', {
        'content': attachment_point.render_js(request=request,
                                              context=context),
    })

    # Now render the HTML for this attachment point.
    try:
        return attachment_point.render(request=request,
                                       context=context)
    except Exception as e:
        logger.exception('Unexpected error rendering actions attachment '
                         'point %r: %s',
                         attachment_point, e,
                         extra={'request': request})

        return mark_safe('')


@register.simple_tag(takes_context=True)
def child_actions_html(
    context: Context,
) -> SafeString:
    """Render all the child actions of the current menu.

    Version Added:
        6.0

    Args:
        context (django.template.Context):
            The current template rendering context.

    Returns:
        django.utils.safestring.SafeString:
        The rendered HTML.

    Raises:
        TypeError:
            The action renderer was an invalid type for this action.
    """
    parent_renderer = context['action_renderer']

    if not isinstance(parent_renderer, BaseActionGroupRenderer):
        raise TypeError(
            f"Attempted to use {{% child_actions_html %}} from a renderer "
            f"that is not a subclass of BaseActionGroupRenderer (it's a "
            f"{type(parent_renderer)!r}"
        )

    return parent_renderer.render_children(
        children=context['children'],
        context=context,
        request=context['request'],
    )


@register.simple_tag(takes_context=True)
def actions_js(
    context: Context,
) -> SafeString:
    """Render the actions JavaScript.

    Version Added:
        6.0

    Args:
        context (django.template.Context):
            The current template rendering context.

    Returns:
        django.utils.safestring.SafeString:
        The rendered JavaScript.
    """
    return mark_safe(''.join(_iter_actions_js(context=context)))


def _iter_actions_js(
    *,
    context: Context,
) -> Iterator[SafeString | str]:
    request = context['request']

    # Render all the action model registration.
    actions: Iterable[BaseAction] = actions_registry

    for action in actions:
        if action.should_register(request=request,
                                  context=context):
            try:
                yield action.render_model_js(request=request,
                                             context=context)
            except Exception as e:
                logger.exception('Error rendering JavaScript for action model '
                                 '%r: %s',
                                 action, e)

    # Render all the action view construction.
    page_state = PageState.for_request(request)

    yield from page_state.iter_content(
        point_name='rb-action-views',
        request=request,
        context=context)


@register.simple_tag(takes_context=True)
def action_js_model_data(
    context: Context,
    action: BaseAction,
) -> SafeString:
    """Render the JS model data for an action.

    Version Added:
        6.0

    Args:
        context (django.template.Context):
            The current template rendering context.

        action (reviewboard.actions.base.BaseAction):
            The action to render.

    Returns:
        django.utils.safestring.SafeString:
        The rendered JavaScript.
    """
    return mark_safe(json.dumps(
        action.get_js_model_data(context=context)))


@register.simple_tag(takes_context=True)
def action_js_view_data_items(
    context: Context,
    action: BaseAction,
) -> SafeString:
    """Render the JS view data for an action.

    Version Added:
        6.0

    Args:
        context (django.template.Context):
            The current template rendering context.

        action (reviewboard.actions.base.BaseAction):
            The action to render.

    Returns:
        django.utils.safestring.SafeString:
        The rendered JavaScript.
    """
    encoded = json.dumps(context['js_view_data'],
                         sort_keys=True)[1:-1]

    if encoded:
        encoded += ','

    return mark_safe(encoded)



"""Unit tests for the DiffCommitResource."""

from django.contrib.auth.models import User
from djblets.features.testing import override_feature_checks
from djblets.testing.decorators import add_fixtures
from djblets.webapi.errors import PERMISSION_DENIED
from djblets.webapi.testing.decorators import webapi_test_template

from reviewboard.reviews.models import ReviewRequest
from reviewboard.webapi.resources import resources
from reviewboard.webapi.tests.base import BaseWebAPITestCase
from reviewboard.webapi.tests.mimetypes import (diffcommit_item_mimetype,
                                                diffcommit_list_mimetype)
from reviewboard.webapi.tests.mixins import (BasicTestsMetaclass,
                                             ReviewRequestChildItemMixin,
                                             ReviewRequestChildListMixin)
from reviewboard.webapi.tests.mixins_extra_data import ExtraDataItemMixin
from reviewboard.webapi.tests.urls import (get_diffcommit_item_url,
                                           get_diffcommit_list_url)


def compare_diffcommit(self, item_rsp, item):
    """Compare a serialized DiffCommit to the original.

    Args:
        item_rsp (dict):
            The serialized response.

        item (reviewboard.diffviewer.models.diffcommit.DiffCommit):
            The DiffCommit to compare against.

    Raises:
        AssertionError:
            The serialized response is not equivalent to the original
            DiffCommit.
    """
    self.assertEqual(item_rsp['id'], item.pk)
    self.assertEqual(item_rsp['commit_id'], item.commit_id)
    self.assertEqual(item_rsp['parent_id'], item.parent_id)
    self.assertEqual(item_rsp['commit_message'], item.commit_message)
    self.assertEqual(item_rsp['author_name'], item.author_name)
    self.assertEqual(item_rsp['author_email'], item.author_email)
    self.assertEqual(item_rsp['committer_name'], item.committer_name)
    self.assertEqual(item_rsp['committer_email'], item.committer_email)


class ResourceListTests(ReviewRequestChildListMixin, BaseWebAPITestCase,
                        metaclass=BasicTestsMetaclass):
    """Tests for DiffCommitResource list resource."""

    fixtures = ['test_users', 'test_scmtools']
    sample_api_url = 'review-request/<id>/diffs/<revision>/commits/'
    resource = resources.diffcommit

    compare_item = compare_diffcommit

    def setup_http_not_allowed_item_test(self, user):
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(
            repository=repository,
            create_with_history=True,
            public=True)

        diffset = self.create_diffset(review_request=review_request)
        return get_diffcommit_list_url(review_request, diffset.revision)

    def setup_http_not_allowed_list_test(self, user):
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(
            repository=repository,
            create_with_history=True,
            public=True)

        diffset = self.create_diffset(review_request=review_request)
        return get_diffcommit_list_url(review_request, diffset.revision)

    def setup_review_request_child_test(self, review_request):
        review_request.extra_data = review_request.extra_data or {}
        review_request.extra_data[
            ReviewRequest._CREATED_WITH_HISTORY_EXTRA_DATA_KEY] = True
        review_request.save(update_fields=('extra_data',))

        diffset = self.create_diffset(review_request=review_request)

        return (get_diffcommit_list_url(review_request,
                                        diffset.revision),
                diffcommit_list_mimetype)

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name,
                             populate_items):
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(
            with_local_site=with_local_site,
            repository=repository,
            public=True)
        diffset = self.create_diffset(review_request=review_request)
        items = []

        if populate_items:
            items.append(self.create_diffcommit(diffset=diffset,
                                                repository=repository))

        return (get_diffcommit_list_url(review_request,
                                        diffset.revision,
                                        local_site_name=local_site_name),
                diffcommit_list_mimetype,
                items)


class ResourceItemTests(ExtraDataItemMixin, ReviewRequestChildItemMixin,
                        BaseWebAPITestCase, metaclass=BasicTestsMetaclass):
    """Tests for DiffCommitResource item resource."""

    fixtures = ['test_users', 'test_scmtools']
    sample_api_url = \
        'review-request/<id>/diffs/<revision>/commits/<commit-id>/'
    resource = resources.diffcommit

    compare_item = compare_diffcommit

    def setup_review_request_child_test(self, review_request):
        diffset = self.create_diffset(review_request=review_request)
        review_request.extra_data[
            ReviewRequest._CREATED_WITH_HISTORY_EXTRA_DATA_KEY] = True
        review_request.save(update_fields=('extra_data',))
        commit = self.create_diffcommit(diffset=diffset,
                                        repository=review_request.repository)

        return (get_diffcommit_item_url(review_request,
                                        diffset.revision,
                                        commit.commit_id),
                diffcommit_item_mimetype)

    def setup_http_not_allowed_item_test(self, user):
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(
            repository=repository,
            public=True)
        diffset = self.create_diffset(review_request=review_request)
        commit = self.create_diffcommit(diffset=diffset,
                                        repository=repository)
        return get_diffcommit_item_url(review_request, diffset.revision,
                                       commit.commit_id)

    def setup_http_not_allowed_list_test(self, user):
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(
            repository=repository,
            submitter=user)
        diffset = self.create_diffset(review_request)
        commit = self.create_diffcommit(diffset=diffset,
                                        repository=repository)

        return get_diffcommit_item_url(review_request,
                                       diffset.revision,
                                       commit.commit_id)

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name):
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(
            repository=repository,
            submitter=user,
            with_local_site=with_local_site)
        diffset = self.create_diffset(review_request)
        commit = self.create_diffcommit(diffset=diffset,
                                        repository=repository)

        return (get_diffcommit_item_url(review_request, diffset.revision,
                                        commit.commit_id, local_site_name),
                diffcommit_item_mimetype,
                commit)

    @webapi_test_template
    def test_get_patch(self):
        """Testing the GET <URL> API with Accept: text/x-patch"""
        url = self.setup_basic_get_test(self.user,
                                        with_local_site=False,
                                        local_site_name=None)[0]

        with override_feature_checks(self.override_features):
            rsp = self.api_get(url,
                               expected_mimetype='text/x-patch',
                               expected_json=False,
                               HTTP_ACCEPT='text/x-patch')

        self.assertEqual(self.DEFAULT_GIT_FILEDIFF_DATA_DIFF, rsp)

    @add_fixtures(['test_site'])
    @webapi_test_template
    def test_get_patch_local_site(self):
        """Testing the GET <URL> API with Accept: text/x-patch on a Local Site
        """
        url = self.setup_basic_get_test(
            User.objects.get(username='doc'),
            with_local_site=True,
            local_site_name=self.local_site_name)[0]

        self.client.login(username='doc', password='doc')

        with override_feature_checks(self.override_features):
            rsp = self.api_get(url,
                               expected_mimetype='text/x-patch',
                               expected_json=False,
                               HTTP_ACCEPT='text/x-patch')

        self.assertEqual(self.DEFAULT_GIT_FILEDIFF_DATA_DIFF, rsp)

    @add_fixtures(['test_site'])
    @webapi_test_template
    def test_get_patch_local_site_no_access(self):
        """Testing the GET <URL> API with Accept: text/x-patch on a Local Site
        without access
        """
        url = self.setup_basic_get_test(
            User.objects.get(username='doc'),
            with_local_site=True,
            local_site_name=self.local_site_name)[0]

        with override_feature_checks(self.override_features):
            rsp = self.api_get(url,
                               expected_status=403,
                               HTTP_ACCEPT='text/x-patch')

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], PERMISSION_DENIED.code)

    @webapi_test_template
    def test_get_patch_private_repository(self):
        """Testing the GET <URL> API with Accept: text/x-patch on a private
        repository
        """
        doc = User.objects.get(username='doc')

        repository = self.create_repository(tool_name='Git', public=False)
        repository.users.add(doc)

        review_request = self.create_review_request(repository=repository,
                                                    submitter=doc)
        diffset = self.create_diffset(review_request)
        commit = self.create_diffcommit(diffset=diffset, repository=repository)

        self.client.login(username='doc', password='doc')

        with override_feature_checks(self.override_features):
            rsp = self.api_get(
                get_diffcommit_item_url(review_request, diffset.revision,
                                        commit.commit_id),
                expected_mimetype='text/x-patch',
                expected_json=False,
                HTTP_ACCEPT='text/x-patch')

        self.assertEqual(self.DEFAULT_GIT_FILEDIFF_DATA_DIFF, rsp)

    @webapi_test_template
    def test_get_patch_private_repository_no_access(self):
        """Testing the GET <URL> API with Accept: text/x-patch on a private
        repository
        """
        doc = User.objects.get(username='doc')

        repository = self.create_repository(tool_name='Git', public=False)
        repository.users.set([doc])

        review_request = self.create_review_request(repository=repository,
                                                    submitter=doc)
        diffset = self.create_diffset(review_request)
        commit = self.create_diffcommit(diffset=diffset, repository=repository)

        with override_feature_checks(self.override_features):
            rsp = self.api_get(
                get_diffcommit_item_url(review_request, diffset.revision,
                                        commit.commit_id),
                expected_status=403)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], PERMISSION_DENIED.code)

    #
    # HTTP PUT tests
    #

    def setup_basic_put_test(self, user, with_local_site, local_site_name,
                             put_valid_data):
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(
            repository=repository,
            submitter=user,
            with_local_site=with_local_site)
        diffset = self.create_diffset(review_request)
        commit = self.create_diffcommit(diffset=diffset,
                                        repository=repository)

        return (get_diffcommit_item_url(review_request,
                                        diffset.revision,
                                        commit.commit_id,
                                        local_site_name=local_site_name),
                diffcommit_item_mimetype,
                {},
                commit,
                [])

    def check_put_result(self, user, item_rsp, item):
        self.compare_item(item_rsp, item)
